from galaxy.tool_util.verify.wait import (
    TimeoutAssertionError,
    wait_on,
)


class Sleeper:
    def __init__(self):
        self.sleeps = []

    def sleep(self, delta):
        self.sleeps.append(delta)


class WaitCondition:
    def __init__(self, after_call_count=0, return_value=True):
        self.after_call_count = after_call_count
        self.call_count = 0
        self.return_value = return_value

    def __call__(self):
        self.call_count += 1
        if self.call_count > self.after_call_count:
            return self.return_value
        else:
            return None


def test_immediate_return():
    condition = WaitCondition(return_value="first")
    sleeper = Sleeper()

    assert "first" == wait_on(condition, "condition", 100, sleep_=sleeper.sleep)
    assert len(sleeper.sleeps) == 0


def test_timeout_first_call():
    condition = WaitCondition(after_call_count=1, return_value="second")
    sleeper = Sleeper()

    exception_called = False
    try:
        wait_on(condition, "never met condition", 1, delta=2, sleep_=sleeper.sleep)
    except TimeoutAssertionError as e:
        assert "never met condition" in str(e)
        assert "after 2.0 seconds" in str(e)
        exception_called = True
    assert len(sleeper.sleeps) == 1
    assert sleeper.sleeps[0] == 2  # delta of 2
    assert exception_called


def test_timeout_third_call():
    condition = WaitCondition(after_call_count=4, return_value="second")
    sleeper = Sleeper()

    exception_called = False
    try:
        wait_on(condition, "condition", 5, delta=2, sleep_=sleeper.sleep)
    except TimeoutAssertionError:
        exception_called = True
    assert len(sleeper.sleeps) == 3
    assert sleeper.sleeps[0] == 2  # delta of 2
    assert sleeper.sleeps[1] == 2  # delta of 2
    assert sleeper.sleeps[2] == 2  # delta of 2
    assert exception_called


def test_return_on_third_call():
    condition = WaitCondition(after_call_count=2, return_value="second")
    sleeper = Sleeper()

    assert "second" == wait_on(condition, "condition", 5, delta=2, sleep_=sleeper.sleep)
    assert len(sleeper.sleeps) == 2
    assert sleeper.sleeps[0] == 2  # delta of 2
    assert sleeper.sleeps[1] == 2  # delta of 2


def test_timeout_backoff():
    condition = WaitCondition(after_call_count=4, return_value="second")
    sleeper = Sleeper()

    exception_called = False
    try:
        wait_on(condition, "condition", 8, delta=2, polling_backoff=1, sleep_=sleeper.sleep)
    except TimeoutAssertionError:
        exception_called = True
    assert len(sleeper.sleeps) == 3
    assert sleeper.sleeps[0] == 2  # delta of 2
    assert sleeper.sleeps[1] == 3  # delta of 2 + 1 backoff
    assert sleeper.sleeps[2] == 4  # delta of 2 + 2 backoff
    assert exception_called



import yaml

from galaxy.tool_util.parser.yaml import to_test_assert_list

# Legacy style
ASSERT_THAT_LIST = yaml.safe_load("""
- that: "has_text"
  text: "Number of input reads |\t1051466"
- that: "has_text"
  text: "Uniquely mapped reads number |\t871202"
""")
# New list of assertion style
ASSERT_LIST = yaml.safe_load("""
- has_text:
    text: "Number of input reads |\t1051466"
- has_text:
    text: "Uniquely mapped reads number |\t871202"
""")
# Singleton assertion
SIMPLE_ASSERT = {"has_text": {"text": "Number of input reads |\t1051466"}}


def test_assert_that_list_to_test_assert_list():
    to_test_assert_list(ASSERT_THAT_LIST)


def test_assert_list_to_test_assert_list():
    to_test_assert_list(ASSERT_LIST)


def test_simple_assert_to_test_assert_list():
    to_test_assert_list(SIMPLE_ASSERT)


def test_assert_legacy_same_as_new_list_style():
    assert to_test_assert_list(ASSERT_THAT_LIST) == to_test_assert_list(ASSERT_THAT_LIST)


NESTED_ASSERT_LIST = yaml.safe_load("""
- has_archive_member:
    path: ".*"
    asserts:
      - has_text:
          text: "a text"
      - has_text:
          text: "another text"
""")


def test_nested_asserts():
    asserts = to_test_assert_list(NESTED_ASSERT_LIST)
    assert asserts and len(asserts) == 1
    assert asserts and asserts[0]["children"] and len(asserts[0]["children"]) == 2
    assert asserts and asserts[0]["children"] and asserts[0]["children"][0]["tag"] == "has_text"
    assert asserts and asserts[0]["children"] and asserts[0]["children"][0]["attributes"]["text"] == "a text"
    assert asserts and asserts[0]["children"] and asserts[0]["children"][1]["tag"] == "has_text"
    assert asserts and asserts[0]["children"] and asserts[0]["children"][1]["attributes"]["text"] == "another text"



import collections
import gzip
import io
import math
import tempfile
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)

import numpy
import pytest
from PIL import Image

from galaxy.tool_util.verify import (
    files_contains,
    files_delta,
    files_diff,
    files_image_diff,
    files_re_match,
    files_re_match_multiline,
)

F1 = b"A\nB\nC"
F2 = b"A\nB\nD\nE" * 61
F3 = b"A\nB\n\xfc"
F4 = b"A\r\nB\nC"
MULTILINE_MATCH = b".*"
TestFile = collections.namedtuple("TestFile", "value path")

TestDef = Tuple[bytes, bytes, Optional[Dict[str, Any]], Optional[Type[AssertionError]]]


def _encode_image(im, **kwargs):
    buf = io.BytesIO()
    pil_im = Image.fromarray(im)
    pil_im.save(buf, **kwargs)
    return buf.getvalue()


F6 = _encode_image(
    numpy.array(
        [
            [255, 255, 255],
            [255, 200, 255],
            [255, 255, 255],
        ],
        dtype=numpy.uint8,
    ),
    format="PNG",
)
F7 = _encode_image(
    numpy.array(
        [
            [255, 255, 255],
            [255, 100, 255],
            [255, 255, 255],
        ],
        dtype=numpy.uint8,
    ),
    format="TIFF",
)
F8 = _encode_image(
    numpy.array(
        [
            [255, 255, 255],
            [255, 100, 255],
            [255, 255, 255],
        ],
        dtype=float,
    )
    / 0xFF,
    format="TIFF",
)
F9 = _encode_image(
    numpy.array(
        [
            [200, 200, 200],
            [200, 1, 200],
            [200, 1, 2],
        ],
        dtype=numpy.uint8,
    ),
    format="PNG",
)
F10 = _encode_image(
    numpy.array(
        [
            [True, True, True],
            [True, False, True],
            [True, True, True],
        ],
        dtype=bool,
    ),
    format="TIFF",
)


def _test_file_list():
    files = []
    for b, ext in [
        (F1, ".txt"),
        (F2, ".txt"),
        (F3, ".pdf"),
        (F4, ".txt"),
        (MULTILINE_MATCH, ".txt"),
        (F1, ".txt.gz"),
        (F6, ".png"),
        (F7, ".tiff"),
        (F8, ".tiff"),
        (F9, ".png"),
        (F10, ".tiff"),
    ]:
        with tempfile.NamedTemporaryFile(mode="wb", suffix=ext, delete=False) as out:
            if ext == ".txt.gz":
                b = gzip.compress(b)
            out.write(b)
        files.append(TestFile(b, out.name))
    return files


def generate_tests(multiline=False):
    f1, f2, f3, f4, multiline_match, f5, f6, f7, f8, f9, f10 = _test_file_list()
    tests: List[TestDef]
    if multiline:
        tests = [(multiline_match, f1, {"lines_diff": 0, "sort": True}, None)]
    else:
        tests = [(f1, f1, {"lines_diff": 0, "sort": True}, None)]
    tests.extend(
        [
            (f1, f2, {"lines_diff": 0, "sort": True}, AssertionError),
            (f1, f3, None, AssertionError),
            (f1, f4, None, None),
            (f1, f5, {"decompress": True}, None),
        ]
    )
    return tests


def generate_tests_sim_size():
    f1, f2, f3, f4, multiline_match, f5, f6, f7, f8, f9, f10 = _test_file_list()
    # tests for equal files
    tests: List[TestDef] = [
        (f1, f1, None, None),  # pass default values
        (f1, f1, {"delta": 0}, None),  # pass for values that should always pass
        (f1, f1, {"delta_frac": 0.0}, None),
    ]
    # tests for two different file (diff is 422 lines, 0.011709602)
    tests += [
        (f1, f2, None, None),  # pass default values
        (f1, f2, {"delta": 422}, None),  # test around the actual difference of two different
        (f1, f2, {"delta_frac": 84.4}, None),
        (f1, f2, {"delta": 421}, AssertionError),
        (f1, f2, {"delta_frac": 84.3}, AssertionError),
        (
            f1,
            f2,
            {"delta": 422, "delta_frac": 84.3},
            AssertionError,
        ),  # test with combination where at least one leads to an assertion
        (f1, f2, {"delta": 421, "delta_frac": 84.4}, AssertionError),
    ]
    return tests


def generate_tests_image_diff():
    f1, f2, f3, f4, multiline_match, f5, f6, f7, f8, f9, f10 = _test_file_list()
    metrics = ["mae", "mse", "rms", "fro", "iou"]
    # tests for equal files (uint8, PNG)
    tests: List[TestDef] = [(f6, f6, {"metric": metric}, None) for metric in metrics]
    # tests for equal files (uint8, TIFF)
    tests += [(f7, f7, {"metric": metric}, None) for metric in metrics]
    # tests for equal files (float, TIFF)
    tests += [(f8, f8, {"metric": metric}, None) for metric in metrics]
    # tests for equal files (bool, TIFF)
    tests += [(f10, f10, {"metric": metric}, None) for metric in metrics]
    # tests for pairs of different files
    tests += [(f6, f8, {"metric": metric}, AssertionError) for metric in metrics]  # uint8 vs float
    tests += [(f7, f8, {"metric": metric}, AssertionError) for metric in metrics]  # uint8 vs float
    tests += [
        (f6, f7, {"metric": "iou"}, None),
        (f6, f7, {"metric": "mae", "eps": 100 / 9 + 1e-4}, None),
        (f6, f7, {"metric": "mae", "eps": 100 / 9 - 1e-4}, AssertionError),
        (f6, f7, {"metric": "mse", "eps": (100**2) / 9 + 1e-4}, None),
        (f6, f7, {"metric": "mse", "eps": (100**2) / 9 - 1e-4}, AssertionError),
        (f6, f7, {"metric": "rms", "eps": math.sqrt((100**2) / 9) + 1e-4}, None),
        (f6, f7, {"metric": "rms", "eps": math.sqrt((100**2) / 9) - 1e-4}, AssertionError),
        (f6, f7, {"metric": "fro", "eps": 100 + 1e-4}, None),
        (f6, f7, {"metric": "fro", "eps": 100 - 1e-4}, AssertionError),
        (f6, f9, {"metric": "iou", "eps": (1 - 1 / 8) + 1e-4}, None),
        (f6, f9, {"metric": "iou", "eps": (1 - 1 / 8) - 1e-4}, AssertionError),
        # tests `pin_labels` with a label not present in any image
        (f6, f9, {"metric": "iou", "eps": 0.999999, "pin_labels": "5"}, AssertionError),
        # tests `pin_labels` with a label present in both images
        (f6, f9, {"metric": "iou", "eps": 0.999999, "pin_labels": "200"}, AssertionError),
        (f6, f9, {"metric": "iou", "eps": 1.0, "pin_labels": "200"}, None),
        # tests `pin_labels` with a label only present in one image
        (f6, f9, {"metric": "iou", "eps": 1.0, "pin_labels": "200, 1"}, AssertionError),
        (f6, f9, {"metric": "iou", "eps": 1.0, "pin_labels": "200, 255"}, AssertionError),
    ]
    return tests


@pytest.mark.parametrize("file1,file2,attributes,expect", generate_tests())
def test_files_contains(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_contains(file1.path, file2.path, attributes)
    else:
        files_contains(file2.path, file2.path, attributes)


@pytest.mark.parametrize("file1,file2,attributes,expect", generate_tests_sim_size())
def test_files_delta(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_delta(file1.path, file2.path, attributes)
    else:
        files_delta(file1.path, file2.path, attributes)


@pytest.mark.parametrize("file1,file2,attributes,expect", generate_tests())
def test_files_diff(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_diff(file1.path, file2.path, attributes)
    else:
        files_diff(file1.path, file2.path, attributes)


@pytest.mark.parametrize("file1,file2,attributes,expect", generate_tests())
def test_files_re_match(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_re_match(file1.path, file2.path, attributes)
    else:
        files_re_match(file1.path, file2.path, attributes)


@pytest.mark.parametrize("file1,file2,attributes,expect", generate_tests(multiline=True))
def test_files_re_match_multiline(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_re_match_multiline(file1.path, file2.path, attributes)
    else:
        files_re_match_multiline(file1.path, file2.path, attributes)


@pytest.mark.parametrize("file1,file2,attributes,expect", generate_tests_image_diff())
def test_files_image_diff(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_image_diff(file1.path, file2.path, attributes)
    else:
        files_image_diff(file1.path, file2.path, attributes)



import json
import os
from tempfile import NamedTemporaryFile
from typing import (
    cast,
    NoReturn,
)
from unittest import mock

from galaxy.tool_util.unittest_utils.interactor import (
    EXISTING_HISTORY,
    EXISTING_HISTORY_NAME,
    EXISTING_SUITE_NAME,
    MockGalaxyInteractor,
    NEW_HISTORY_ID,
)
from galaxy.tool_util.verify.interactor import GalaxyInteractorApi
from galaxy.tool_util.verify.script import (
    arg_parser,
    build_case_references,
    Results,
    test_tools as run,
    TestReference,
)

VT_PATH = "galaxy.tool_util.verify.script.verify_tool"


def test_arg_parse() -> None:
    parser = arg_parser()

    # defaults
    args = parser.parse_args([])
    assert not args.with_reference_data
    assert not args.history_per_test_case
    assert args.page_size == 0
    assert args.page_number == 0
    assert args.parallel_tests == 1

    # skip flags
    args = parser.parse_args(["--skip-with-reference-data", "--history-per-test-case"])
    assert not args.with_reference_data
    assert args.history_per_test_case

    # enable flags
    args = parser.parse_args(["--with-reference-data", "--history-per-suite"])
    assert args.with_reference_data
    assert not args.history_per_test_case

    # pagination
    args = parser.parse_args(["--page-size", "5", "--page-number", "40", "--parallel-tests", "3"])
    assert args.page_size == 5
    assert args.page_number == 40
    assert args.parallel_tests == 3
    assert args.test_data is None

    args = parser.parse_args(["--test-data", "/foo/bar", "--test-data", "cow"])
    assert len(args.test_data) == 2
    assert args.test_data[0] == "/foo/bar"
    assert args.test_data[1] == "cow"

    args = parser.parse_args(["--skip-previously-successful"])
    assert args.skip == "successful"

    args = parser.parse_args(["--skip-previously-executed"])
    assert args.skip == "executed"


def test_test_tools() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results("my suite", f.name)
    test_references = [
        TestReference("cat", "0.1.0", 0),
        TestReference("cat", "0.1.0", 1),
        TestReference("cat", "0.2.0", 0),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 3
        assert len(results.test_exceptions) == 0
        assert_results_written(results)
        assert interactor.history_created
        assert interactor.history_deleted


def test_test_tools_no_history_cleanup() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results("my suite", f.name)
    test_references = [
        TestReference("cat", "0.1.0", 0),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
            no_history_cleanup=True,
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 1
        assert len(results.test_exceptions) == 0
        assert_results_written(results)
        assert interactor.history_created
        assert not interactor.history_deleted


def test_test_tools_history_reuse() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results(EXISTING_SUITE_NAME, f.name)
    test_references = [
        TestReference("cat", "0.1.0", 0),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
            no_history_reuse=False,
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 1
        assert len(results.test_exceptions) == 0
        assert_results_written(results)
        assert interactor.history_id == EXISTING_HISTORY["id"]
        assert interactor.history_name == EXISTING_HISTORY_NAME
        assert not interactor.history_created
        assert not interactor.history_deleted


def test_test_tools_no_history_reuse() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results("existing suite", f.name)
    test_references = [
        TestReference("cat", "0.1.0", 0),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
            no_history_reuse=True,
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 1
        assert len(results.test_exceptions) == 0
        assert_results_written(results)
        assert interactor.history_id == NEW_HISTORY_ID
        assert interactor.history_name == EXISTING_HISTORY_NAME
        assert interactor.history_created
        assert interactor.history_deleted


def test_test_tools_history_name() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results("my suite", f.name)
    test_references = [
        TestReference("cat", "0.1.0", 0),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
            history_name="testfoo",
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 1
        assert len(results.test_exceptions) == 0
        assert_results_written(results)
        assert interactor.history_id == NEW_HISTORY_ID
        assert interactor.history_name == "testfoo"
        assert interactor.history_created
        assert interactor.history_deleted


def test_test_tool_per_test_history() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results("my suite", f.name)
    test_references = [
        TestReference("cat", "0.1.0", 0),
        TestReference("cat", "0.1.0", 1),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
            history_per_test_case=True,
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 2
        assert len(results.test_exceptions) == 0
        assert_results_written(results)
        assert not interactor.history_created
        assert not interactor.history_deleted


def test_test_tools_records_exception() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results("my suite", f.name)
    test_references = [
        TestReference("bad", "0.1.0", 0),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)

        def side_effect(*args, **kwd) -> NoReturn:
            raise Exception("Cow")

        mock_verify.side_effect = side_effect
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 1
        assert len(results.test_exceptions) == 1
        assert_results_written(results)


def test_test_tools_records_retry_exception() -> None:
    interactor = MockGalaxyInteractor()
    f = NamedTemporaryFile()
    results = Results("my suite", f.name)
    test_references = [
        TestReference("bad", "0.1.0", 0),
    ]
    with mock.patch(VT_PATH) as mock_verify:
        assert_results_not_written(results)

        count = 0

        def side_effect(*args, **kwd):
            nonlocal count
            raise_exception = count == 0
            count += 1
            if raise_exception:
                raise Exception("Cow")

        mock_verify.side_effect = side_effect
        run(
            cast(GalaxyInteractorApi, interactor),
            test_references,
            results,
            retries=1,
        )
        calls = mock_verify.call_args_list
        assert len(calls) == 2
        assert len(results.test_exceptions) == 0
        assert_results_written(results)


def test_append_results() -> None:
    tf = NamedTemporaryFile()
    tf.close()
    assert not os.path.exists(tf.name)
    # Now try to append to the non-existent file.
    results = Results("my suite", tf.name, True)
    results.register_result({"id": "foo", "has_data": True, "data": {"status": "success"}})
    try:
        results.write()
        assert os.path.exists(tf.name)
        with open(tf.name) as f:
            report_obj = json.load(f)
        assert "tests" in report_obj
        assert len(report_obj["tests"]) == 1
        assert report_obj["results"]["total"] == 1, report_obj["results"]
        assert report_obj["results"]["errors"] == 0
        assert report_obj["results"]["skips"] == 0
    finally:
        os.remove(tf.name)
    assert not os.path.exists(tf.name)


def test_results():
    tf = NamedTemporaryFile()
    results = Results("my suite", tf.name)
    results.register_result({"id": "foo", "has_data": True, "data": {"status": "success"}})
    results.write()
    message = results.info_message()

    with open(tf.name) as f:
        report_obj = json.load(f)
    assert "tests" in report_obj
    assert len(report_obj["tests"]) == 1
    assert report_obj["results"]["total"] == 1, report_obj["results"]
    assert report_obj["results"]["errors"] == 0
    assert report_obj["results"]["skips"] == 0

    assert "Passed tool tests (1)" in message
    assert "Skipped tool tests (0)" in message

    results.register_result({"id": "bar", "has_data": True, "data": {"status": "skip"}})
    results.write()
    message = results.info_message()

    with open(tf.name) as f:
        report_obj = json.load(f)
    assert len(report_obj["tests"]) == 2
    assert report_obj["results"]["skips"] == 1
    assert "Passed tool tests (1)" in message
    assert "Skipped tool tests (1)" in message


def test_build_references() -> None:
    interactor = cast(GalaxyInteractorApi, MockGalaxyInteractor())
    test_references = build_case_references(interactor)
    assert len(test_references) == 6

    test_references = build_case_references(interactor, "cat1", tool_version="*")
    assert len(test_references) == 6

    test_references = build_case_references(interactor, "cat1", tool_version="*", page_size=3, page_number=1)
    assert len(test_references) == 3

    test_references = build_case_references(interactor, "cat1", tool_version="*", page_size=3, page_number=2)
    assert len(test_references) == 0

    test_references = build_case_references(interactor, "cat1", tool_version="*", page_size=3, page_number=3)
    assert len(test_references) == 0

    test_references = build_case_references(interactor, "cat1", tool_version=None)
    assert len(test_references) == 4

    test_references = build_case_references(interactor, "cat1", tool_version="0.2.0")
    assert len(test_references) == 4

    test_references = build_case_references(interactor, "cat1", tool_version="0.1.0")
    assert len(test_references) == 2

    test_references = build_case_references(interactor, "cat1", tool_version="0.1.0", test_index=1)
    assert len(test_references) == 1

    # Specifying an index but not a version, grabs latest version and fills it in.
    test_references = build_case_references(interactor, "cat1", test_index=2)
    assert len(test_references) == 1
    test_reference = test_references[0]
    assert test_reference.tool_id == "cat1"
    assert test_reference.tool_version == "0.2.0"
    assert test_reference.test_index == 2


def assert_results_not_written(results: Results) -> None:
    assert os.stat(results.test_json).st_size == 0


def assert_results_written(results: Results) -> None:
    assert os.stat(results.test_json).st_size > 0
    with open(results.test_json) as f:
        json.load(f)



from os import environ
from typing import Dict

import pytest

from .util import modify_environ


@pytest.fixture
def load_keyval(request):
    """
    Create key/value pair and load it into os.environ. Delete on teardown.
    """
    keys = []  # preserve keys for teardown

    def _load_keyval(key="a unique key", val="a value"):
        # If this is called twice with default values within the same test function,
        #   it will raise a KeyError. This is intentional: os.environ cannot have duplicate keys.
        keys.append(key)
        environ[key] = val
        return key, val

    def _teardown():
        for k in keys:
            del environ[k]

    request.addfinalizer(_teardown)
    return _load_keyval


def test_modify_environ__restore(load_keyval):
    key, val = load_keyval()
    with modify_environ({}):
        assert environ[key] == val  # key/val unchanged
    assert environ[key] == val  # key/val unchanged


def test_modify_environ__add_and_restore(load_keyval):
    key1, val1 = load_keyval()
    key2, val2 = "key to add", "value to add"
    to_update = {key2: val2}

    assert key2 not in environ  # ensure key to add does not exist
    with modify_environ(to_update):
        assert environ[key1] == val1  # key/val unchanged
        assert environ[key2] == val2  # new key/val added
    assert environ[key1] == val1  # key/val unchanged
    assert key2 not in environ  # new key removed


def test_modify_environ__update_and_restore(load_keyval):
    key1, val1 = load_keyval()
    key2, val2 = load_keyval("key to update", "value to update")
    val2_updated = "updated"
    to_update = {key2: val2_updated}

    with modify_environ(to_update):
        assert environ[key1] == val1  # key/val unchanged
        assert environ[key2] == val2_updated  # value updated
    assert environ[key1] == val1  # key/val unchanged
    assert environ[key2] == val2  # value restored


def test_modify_environ__remove_and_restore(load_keyval):
    key1, val1 = load_keyval()
    key2, val2 = load_keyval("key to remove", "value to remove")
    to_update: Dict[str, str] = {}
    to_remove = [key2]

    with modify_environ(to_update, to_remove):
        assert environ[key1] == val1  # key/val unchanged
        assert key2 not in environ  # key removed
    assert environ[key1] == val1  # key/val unchanged
    assert environ[key2] == val2  # key/value restored


def test_modify_environ__remove_nonexistant_key(load_keyval):
    # Test that removing wrong key does not raise an error
    key1, val1 = load_keyval()
    key_nonexistant = "no such key"
    to_update: Dict[str, str] = {}
    to_remove = [key_nonexistant]

    assert key_nonexistant not in environ  # ensure key to remove does not exist
    with modify_environ(to_update, to_remove):
        assert environ[key1] == val1  # key/val unchanged
    assert environ[key1] == val1  # key/val unchanged



from functools import partial
from typing import (
    Callable,
    List,
    Optional,
)

import yaml

from galaxy.exceptions import RequestParameterInvalidException
from galaxy.tool_util.parameters import (
    decode,
    encode,
    RawStateDict,
    RequestInternalToolState,
    RequestToolState,
    ToolParameterBundleModel,
    validate_internal_job,
    validate_internal_landing_request,
    validate_internal_request,
    validate_internal_request_dereferenced,
    validate_landing_request,
    validate_relaxed_request,
    validate_request,
    validate_test_case,
    validate_workflow_step,
    validate_workflow_step_linked,
    ValidationFunctionT,
)
from galaxy.tool_util.parameters.json import (
    to_json_schema,
    to_json_schema_string,
)
from galaxy.tool_util.unittest_utils.parameters import (
    parameter_bundle_for_file,
    parameter_bundle_for_framework_tool,
)
from galaxy.util.resources import resource_string


def specification_object():
    try:
        yaml_str = resource_string(__name__, "parameter_specification.yml")
    except AttributeError:
        # hack for the main() function below where this file is interpreted as part of the
        # Galaxy tree.
        yaml_str = open("test/unit/tool_util/parameter_specification.yml").read()
    return yaml.safe_load(yaml_str)


def framework_tool_checks():
    # Extend parameter_specification with some extra tests against existing framework tools.
    # There is something beautiful about a targeted tool for every parameter feature but realistically
    # we've been doing a version of the for a decade with tool tests and we can leverage those also.
    try:
        yaml_str = resource_string(__name__, "framework_tool_checks.yml")
    except AttributeError:
        # hack for the main() function below where this file is interpreted as part of the
        # Galaxy tree.
        yaml_str = open("test/unit/tool_util/framework_tool_checks.yml").read()
    return yaml.safe_load(yaml_str)


def test_specification():
    parameter_spec = specification_object()
    for file in parameter_spec.keys():
        _test_file(file, parameter_spec)


def test_framework_tool_checks():
    parameter_spec = framework_tool_checks()
    for file in parameter_spec.keys():
        _test_file(file, parameter_spec, parameter_bundle_for_framework_tool(f"{file}.xml"))


def test_single():
    # _test_file("gx_int")
    # _test_file("gx_float")
    # _test_file("gx_boolean")
    # _test_file("gx_int_optional")
    # _test_file("gx_float_optional")
    # _test_file("gx_conditional_boolean")
    # _test_file("gx_conditional_conditional_boolean")
    _test_file("gx_conditional_boolean_checked")


def _test_file(file: str, specification=None, parameter_bundle: Optional[ToolParameterBundleModel] = None):
    spec = specification or specification_object()
    combos = spec[file]
    if parameter_bundle is None:
        parameter_bundle = parameter_bundle_for_file(file)
    assert parameter_bundle

    assertion_functions = {
        "relaxed_request_valid": _assert_relaxed_requests_validate,
        "relaxed_request_invalid": _assert_relaxed_requests_invalid,
        "request_valid": _assert_requests_validate,
        "request_invalid": _assert_requests_invalid,
        "request_internal_valid": _assert_internal_requests_validate,
        "request_internal_invalid": _assert_internal_requests_invalid,
        "request_internal_dereferenced_valid": _assert_internal_requests_dereferenced_validate,
        "request_internal_dereferenced_invalid": _assert_internal_requests_dereferenced_invalid,
        "landing_request_valid": _assert_landing_requests_validate,
        "landing_request_invalid": _assert_landing_requests_invalid,
        "landing_request_internal_valid": _assert_internal_landing_requests_validate,
        "landing_request_internal_invalid": _assert_internal_landing_requests_invalid,
        "job_internal_valid": _assert_internal_jobs_validate,
        "job_internal_invalid": _assert_internal_jobs_invalid,
        "test_case_xml_valid": _assert_test_cases_validate,
        "test_case_xml_invalid": _assert_test_cases_invalid,
        "workflow_step_valid": _assert_workflow_steps_validate,
        "workflow_step_invalid": _assert_workflow_steps_invalid,
        "workflow_step_linked_valid": _assert_workflow_steps_linked_validate,
        "workflow_step_linked_invalid": _assert_workflow_steps_linked_invalid,
    }

    for valid_or_invalid, tests in combos.items():
        assertion_function = assertion_functions[valid_or_invalid]
        assertion_function(parameter_bundle, tests)

    # Assume request validation will work here.
    if "request_internal_valid" not in combos and "request_valid" in combos:
        _assert_internal_requests_validate(parameter_bundle, combos["request_valid"])
    if "request_internal_invalid" not in combos and "request_invalid" in combos:
        _assert_internal_requests_invalid(parameter_bundle, combos["request_invalid"])


def _for_each(test: Callable, parameters: ToolParameterBundleModel, requests: List[RawStateDict]) -> None:
    for request in requests:
        test(parameters, request)


def model_assertion_function_factory(validate_function: ValidationFunctionT, what: str):

    def _assert_validates(parameters: ToolParameterBundleModel, request: RawStateDict) -> None:
        try:
            validate_function(parameters, request)
        except RequestParameterInvalidException as e:
            raise AssertionError(f"Parameters {parameters} failed to validate {what} {request}. {e}")

    def _assert_invalid(parameters: ToolParameterBundleModel, request: RawStateDict) -> None:
        exc = None
        try:
            validate_function(parameters, request)
        except RequestParameterInvalidException as e:
            exc = e

        if exc is None:
            raise AssertionError(
                f"Parameters {parameters} didn't result in validation error on {what} {request} as expected."
            )

    return _assert_validates, _assert_invalid


_assert_relaxed_request_validates, _assert_relaxed_request_invalid = model_assertion_function_factory(
    validate_relaxed_request, "relaxed request"
)
_assert_request_validates, _assert_request_invalid = model_assertion_function_factory(validate_request, "request")
_assert_internal_request_validates, _assert_internal_request_invalid = model_assertion_function_factory(
    validate_internal_request, "internal request"
)
_assert_internal_request_dereferenced_validates, _assert_internal_request_dereferenced_invalid = (
    model_assertion_function_factory(validate_internal_request_dereferenced, "dereferenced internal request")
)
_assert_internal_job_validates, _assert_internal_job_invalid = model_assertion_function_factory(
    validate_internal_job, "internal job description"
)
_assert_test_case_validates, _assert_test_case_invalid = model_assertion_function_factory(
    validate_test_case, "XML derived test case"
)
_assert_workflow_step_validates, _assert_workflow_step_invalid = model_assertion_function_factory(
    validate_workflow_step, "workflow step tool state (unlinked)"
)
_assert_workflow_step_linked_validates, _assert_workflow_step_linked_invalid = model_assertion_function_factory(
    validate_workflow_step_linked, "linked workflow step tool state"
)
_assert_landing_request_validates, _assert_landing_request_invalid = model_assertion_function_factory(
    validate_landing_request, "landing request"
)
_assert_internal_landing_request_validates, _assert_internal_landing_request_invalid = model_assertion_function_factory(
    validate_internal_landing_request, " internallanding request"
)

_assert_relaxed_requests_validate = partial(_for_each, _assert_relaxed_request_validates)
_assert_relaxed_requests_invalid = partial(_for_each, _assert_relaxed_request_invalid)
_assert_requests_validate = partial(_for_each, _assert_request_validates)
_assert_requests_invalid = partial(_for_each, _assert_request_invalid)
_assert_internal_requests_validate = partial(_for_each, _assert_internal_request_validates)
_assert_internal_requests_invalid = partial(_for_each, _assert_internal_request_invalid)
_assert_internal_requests_dereferenced_validate = partial(_for_each, _assert_internal_request_dereferenced_validates)
_assert_internal_requests_dereferenced_invalid = partial(_for_each, _assert_internal_request_dereferenced_invalid)
_assert_internal_jobs_validate = partial(_for_each, _assert_internal_job_validates)
_assert_internal_jobs_invalid = partial(_for_each, _assert_internal_job_invalid)
_assert_test_cases_validate = partial(_for_each, _assert_test_case_validates)
_assert_test_cases_invalid = partial(_for_each, _assert_test_case_invalid)
_assert_workflow_steps_validate = partial(_for_each, _assert_workflow_step_validates)
_assert_workflow_steps_invalid = partial(_for_each, _assert_workflow_step_invalid)
_assert_workflow_steps_linked_validate = partial(_for_each, _assert_workflow_step_linked_validates)
_assert_workflow_steps_linked_invalid = partial(_for_each, _assert_workflow_step_linked_invalid)
_assert_landing_requests_validate = partial(_for_each, _assert_landing_request_validates)
_assert_landing_requests_invalid = partial(_for_each, _assert_landing_request_invalid)
_assert_internal_landing_requests_validate = partial(_for_each, _assert_internal_landing_request_validates)
_assert_internal_landing_requests_invalid = partial(_for_each, _assert_internal_landing_request_invalid)


def decode_val(val: str) -> int:
    assert val == "abcdabcd"
    return 5


def test_decode_gx_data():
    input_bundle = parameter_bundle_for_file("gx_data")

    request_tool_state = RequestToolState({"parameter": {"src": "hda", "id": "abcdabcd"}})
    request_internal_tool_state = decode(request_tool_state, input_bundle, decode_val)
    assert request_internal_tool_state.input_state["parameter"]["id"] == 5
    assert request_internal_tool_state.input_state["parameter"]["src"] == "hda"


def test_decode_gx_int():
    input_bundle = parameter_bundle_for_file("gx_int")

    request_tool_state = RequestToolState({"parameter": 5})
    request_internal_tool_state = decode(request_tool_state, input_bundle, decode_val)
    assert request_internal_tool_state.input_state["parameter"] == 5


def test_json_schema_for_conditional():
    input_bundle = parameter_bundle_for_file("gx_conditional_boolean")
    tool_state = RequestToolState.parameter_model_for(input_bundle)
    json_schema = to_json_schema(tool_state)
    assert json_schema["$defs"]["When_test_parameter_False"]["properties"]["test_parameter"]["const"] is False
    assert json_schema["$defs"]["When_test_parameter_True"]["properties"]["test_parameter"]["const"] is True
    assert "test_parameter" not in json_schema["$defs"]["When_test_parameter___absent"]["properties"]
    assert to_json_schema_string(tool_state)


def test_encode_gx_data():
    input_bundle = parameter_bundle_for_file("gx_data")

    def encode_val(val: int) -> str:
        assert val == 5
        return "abcdabcd"

    request_internal_tool_state = RequestInternalToolState({"parameter": {"src": "hda", "id": 5}})
    request_tool_state = encode(request_internal_tool_state, input_bundle, encode_val)
    assert request_tool_state.input_state["parameter"]["id"] == "abcdabcd"
    assert request_tool_state.input_state["parameter"]["src"] == "hda"


if __name__ == "__main__":
    parameter_spec = specification_object()
    parameter_models_json = {}
    for file in parameter_spec.keys():
        tool_parameter_model = parameter_bundle_for_file(file)
        parameter_models_json[file] = tool_parameter_model.model_dump()
    yaml_str = yaml.safe_dump(parameter_models_json)
    with open("client/src/components/Tool/parameter_models.yml", "w") as f:
        f.write("# auto generated file for JavaScript testing, do not modify manually\n")
        f.write("# -----\n")
        f.write('# PYTHONPATH="lib" python test/unit/tool_util/test_parameter_specification.py\n')
        f.write("# -----\n")
        f.write(yaml_str)



import csv
import json
import os
from dataclasses import dataclass

from galaxy.tool_util.data import (
    BUNDLE_INDEX_FILE_NAME,
    BundleProcessingOptions,
)
from galaxy.tool_util.data.bundles.models import (
    convert_data_tables_xml,
    DataTableBundleProcessorDescription,
)
from galaxy.util import (
    parse_xml,
)
from galaxy.util.resources import (
    as_file,
    resource_path,
)


def _functional_test_tool_directory() -> str:
    root_dir = os.getcwd()
    return os.path.join(root_dir, "test/functional/tools")


TOOLS_DIRECTORY = os.path.abspath(_functional_test_tool_directory())


def test_xml_parsing() -> None:
    path = os.path.join(TOOLS_DIRECTORY, "sample_data_manager_conf.xml")
    tree = parse_xml(path)
    data_managers_el = tree.getroot()
    data_manager_el = data_managers_el.find("data_manager")
    assert data_manager_el is not None
    description = convert_data_tables_xml(data_manager_el)
    assert not description.undeclared_tables
    assert len(description.data_tables) == 1
    data_table = description.data_tables[0]
    output = data_table.output
    assert output
    columns = output.columns
    assert len(columns) == 2
    column1 = columns[0]
    assert column1.name == "value"
    assert column1.output_ref is None
    column2 = columns[1]
    assert column2.name == "path"
    assert column2.output_ref == "out_file"
    moves = column2.moves
    assert len(moves) == 1
    move = moves[0]
    assert move.type == "directory"
    assert move.relativize_symlinks is True
    assert move.target_base == "${GALAXY_DATA_MANAGER_DATA_PATH}"
    assert move.target_value == "testbeta/${value}"
    assert move.source_base is None
    assert move.source_value == ""


def test_parsing_manual() -> None:
    with as_file(resource_path(__name__, "example_data_managers/manual.xml")) as path:
        tree = parse_xml(path)
    data_managers_el = tree.getroot()
    data_manager_el = data_managers_el.find("data_manager")
    assert data_manager_el is not None
    description = convert_data_tables_xml(data_manager_el)
    assert description.undeclared_tables
    assert len(description.data_tables) == 0


def test_parsing_mothur() -> None:
    with as_file(resource_path(__name__, "example_data_managers/mothur.xml")) as path:
        tree = parse_xml(path)
    data_managers_el = tree.getroot()
    data_manager_el = data_managers_el.find("data_manager")
    assert data_manager_el is not None
    description = convert_data_tables_xml(data_manager_el)
    assert not description.undeclared_tables
    assert len(description.data_tables) == 4


@dataclass
class OutputDataset:
    file_name_: str
    extra_files_path: str
    ext: str = "data_manager_json"

    def get_file_name(self, sync_cache=True) -> str:
        return self.file_name_

    def extra_files_path_exists(self) -> bool:
        return os.path.exists(self.extra_files_path)


def prepare_typical_output_and_description(tmp_path):
    target_path = tmp_path / "newvalue.txt"
    target_path.write_text("Moo Cow")
    output = {"data_tables": {"testalpha": [{"value": "newvalue", "name": "mynewname", "path": "newvalue.txt"}]}}
    output_dataset_path = tmp_path / "output.dat"
    output_dataset_path.write_text(json.dumps(output))
    extra_files_path = tmp_path / "extra"
    extra_files_path.mkdir()
    output_dataset = OutputDataset(
        output_dataset_path,
        extra_files_path,
    )
    out_data = {"out1": output_dataset}
    data_table = {
        "name": "testalpha",
        "output": {
            "columns": [
                {
                    "name": "value",
                },
                {
                    "name": "name",
                },
                {
                    "name": "path",
                    "data_table_name": "path",
                    "output_ref": "out1",
                    "moves": [
                        {
                            "type": "directory",
                            "relativize_symlinks": True,
                            "target_value": "testalpha/${value}",
                            "target_base": "${GALAXY_DATA_MANAGER_DATA_PATH}",
                        }
                    ],
                    "value_translations": [
                        {"value": "${GALAXY_DATA_MANAGER_DATA_PATH}/testalpha/${value}/${path}", "type": "template"},
                        {"value": "abspath", "type": "function"},
                    ],
                },
            ]
        },
    }
    process_description = DataTableBundleProcessorDescription(
        **{
            "undeclared_tables": False,
            "data_tables": [data_table],
        }
    )
    return out_data, process_description


def test_typical_processing(tdt_manager, tmp_path):
    options = BundleProcessingOptions(
        what="data manager 'mock'",
        data_manager_path=str(tmp_path),
        target_config_file=str(tmp_path / "sample_data_managers_conf.xml"),
    )
    out_data, process_description = prepare_typical_output_and_description(tmp_path)
    repo_info = None
    tdt_manager.process_bundle(
        out_data,
        process_description,
        repo_info,
        options,
    )
    loc1 = tmp_path / "testalpha.loc"
    new_row = _last_row(loc1)
    assert new_row[0] == "newvalue"
    assert new_row[1] == "mynewname"
    assert new_row[2] == str(tmp_path / "testalpha" / "newvalue" / "newvalue.txt")


def test_write_bundle(tdt_manager, tmp_path):
    out_data, process_description = prepare_typical_output_and_description(tmp_path)
    tdt_manager.write_bundle(
        out_data,
        process_description,
        repo_info=None,
    )
    extra = tmp_path / "extra"
    bundle_index_json_path = extra / BUNDLE_INDEX_FILE_NAME
    assert bundle_index_json_path.exists()
    with open(bundle_index_json_path) as f:
        bundle_index = json.load(f)
    assert "processor_description" in bundle_index
    assert "data_tables" in bundle_index
    assert "output_name" in bundle_index


def test_import_bundle(tdt_manager, tmp_path):
    out_data, process_description = prepare_typical_output_and_description(tmp_path)
    tdt_manager.write_bundle(
        out_data,
        process_description,
        None,
    )

    # Writing the bundle didn't update the loc files.
    loc1 = tmp_path / "testalpha.loc"
    new_row = _last_row(loc1)
    assert new_row[0] != "newvalue"

    options = BundleProcessingOptions(
        what="data manager 'mock'",
        data_manager_path=str(tmp_path),
        target_config_file=str(tmp_path / "sample_data_managers_conf.xml"),
    )
    tdt_manager.import_bundle(
        str(tmp_path / "extra"),
        options,
    )

    # Importing the bundle does update the loc files though.
    loc1 = tmp_path / "testalpha.loc"
    new_row = _last_row(loc1)
    assert new_row[0] == "newvalue"
    assert new_row[1] == "mynewname"
    assert new_row[2] == str(tmp_path / "testalpha" / "newvalue" / "newvalue.txt")


def test_undeclared_tables(tdt_manager, tmp_path):
    options = BundleProcessingOptions(
        what="data manager 'mock'",
        data_manager_path=str(tmp_path),
        target_config_file=str(tmp_path / "sample_data_managers_conf.xml"),
    )
    target_path = tmp_path / "newvalue.txt"
    target_path.write_text("Moo Cow")
    output = {"data_tables": {"testalpha": [{"value": "newvalue", "name": "mynewname", "path": "newvalue.txt"}]}}
    output_dataset_path = tmp_path / "output.dat"

    output_dataset_path.write_text(json.dumps(output))
    extra_files_path = tmp_path / "extra"
    output_dataset = OutputDataset(
        output_dataset_path,
        extra_files_path,
    )
    out_data = {"out1": output_dataset}
    process_description = DataTableBundleProcessorDescription(
        **{
            "undeclared_tables": True,
            "data_tables": [],
        }
    )
    repo_info = None
    tdt_manager.process_bundle(
        out_data,
        process_description,
        repo_info,
        options,
    )
    loc1 = tmp_path / "testalpha.loc"
    new_row = _last_row(loc1)
    assert new_row[0] == "newvalue"
    assert new_row[1] == "mynewname"
    assert new_row[2] == str(tmp_path / "newvalue.txt")


def _last_row(loc_file):
    with open(loc_file) as file:
        rows = csv.reader(file, delimiter="\t")
        for row in rows:
            last_row = row
        return last_row



import os
from pathlib import Path
from typing import List

import yaml

from galaxy.tool_util_models import Tests
from galaxy.util import galaxy_directory
from galaxy.util.unittest_utils import skip_unless_environ

TEST_WORKFLOW_DIRECTORY = os.path.join(galaxy_directory(), "lib", "galaxy_test", "workflow")
IWC_WORKFLOWS_USING_UNVERIFIED_SYNTAX: List[str] = []


def test_validate_workflow_tests():
    path = Path(TEST_WORKFLOW_DIRECTORY)
    test_files = path.glob("*.gxwf-tests.yml")
    for test_file in test_files:
        with open(test_file) as f:
            json = yaml.safe_load(f)
        Tests.model_validate(json)


@skip_unless_environ("GALAXY_TEST_IWC_DIRECTORY")
def test_iwc_directory():
    path = Path(os.environ["GALAXY_TEST_IWC_DIRECTORY"])
    test_files = path.glob("workflows/**/*-test*.yml")

    for test_file in test_files:
        print(test_file)
        skip_file = False
        for unverified in IWC_WORKFLOWS_USING_UNVERIFIED_SYNTAX:
            if str(test_file).endswith(unverified):
                skip_file = True
        if skip_file:
            continue
        with open(test_file) as f:
            json = yaml.safe_load(f)
        Tests.model_validate(json)



import os

from galaxy.tool_util.deps.mulled.get_tests import main_test_search
from galaxy.tool_util.deps.mulled.mulled_update_singularity_containers import (
    docker_to_singularity,
    get_list_from_file,
    singularity_container_test,
)
from galaxy.util.unittest_utils import skip_unless_executable
from ..util import external_dependency_management


def test_get_list_from_file(tmp_path) -> None:
    list_file = os.path.join(tmp_path, "list_file.txt")
    with open(list_file, "w") as f:
        f.write("bbmap:36.84--0\nbiobambam:2.0.42--0\nconnor:0.5.1--py35_0\ndiamond:0.8.26--0\nedd:1.1.18--py27_0")
    assert get_list_from_file(list_file) == [
        "bbmap:36.84--0",
        "biobambam:2.0.42--0",
        "connor:0.5.1--py35_0",
        "diamond:0.8.26--0",
        "edd:1.1.18--py27_0",
    ]


@external_dependency_management
@skip_unless_executable("singularity")
def test_docker_to_singularity(tmp_path) -> None:
    docker_to_singularity("abundancebin:1.0.1--0", "singularity", tmp_path, no_sudo=True)
    assert tmp_path.joinpath("abundancebin:1.0.1--0").exists()


@external_dependency_management
@skip_unless_executable("singularity")
def test_singularity_container_test(tmp_path) -> None:
    containers = [
        "pybigwig:0.3.22--py36h54a71a5_0",  # test Python imports
        "samtools:1.0--1",
        "yasm:1.3.0--0",  # test missing tests
    ]
    tests = {container: main_test_search(container) for container in containers}
    for n in tests.keys():
        docker_to_singularity(n, "singularity", tmp_path, no_sudo=True)
    results = singularity_container_test(
        tests,
        "singularity",
        tmp_path,
    )
    assert "samtools:1.0--1" in results["passed"], results
    assert "pybigwig:0.3.22--py36h54a71a5_0" in results["passed"], results
    assert "yasm:1.3.0--0" in results["notest"], results



import tempfile
import time
from contextlib import contextmanager
from os import path
from shutil import rmtree

import pytest

from galaxy.tool_util.toolbox import watcher
from galaxy.util import bunch
from galaxy.util.watcher import can_watch


@pytest.mark.skipif(not can_watch, reason="watchdog not available")
def test_watcher():
    with __test_directory() as t:
        tool_path = path.join(t, "test.xml")
        toolbox = Toolbox()
        with open(tool_path, "w") as f:
            f.write("a")
        tool_watcher = watcher.get_tool_watcher(toolbox, bunch.Bunch(watch_tools=True))
        tool_watcher.start()
        tool_watcher.watch_file(tool_path, "cool_tool")
        time.sleep(2)
        assert not toolbox.was_reloaded("cool_tool")
        with open(tool_path, "w") as f:
            f.write("b")
        wait_for_reload(lambda: toolbox.was_reloaded("cool_tool"))
        tool_watcher.shutdown()
        assert tool_watcher.observer is None


@pytest.mark.skipif(not can_watch, reason="watchdog not available")
def test_tool_conf_watcher():
    callback = CallbackRecorder()
    conf_watcher = watcher.get_tool_conf_watcher(callback.call)
    conf_watcher.start()

    with __test_directory() as t:
        tool_conf_path = path.join(t, "test_conf.xml")
        with open(tool_conf_path, "w") as f:
            f.write("a")
        conf_watcher.watch_file(tool_conf_path)
        time.sleep(2)
        with open(tool_conf_path, "w") as f:
            f.write("b")
        wait_for_reload(lambda: callback.called)
        conf_watcher.shutdown()
        assert conf_watcher.thread is None


def wait_for_reload(check):
    reloaded = False
    for _ in range(10):
        reloaded = check()
        if reloaded:
            break
        time.sleep(0.2)
    assert reloaded


class Toolbox:
    def __init__(self):
        self.reloaded = {}

    def reload_tool_by_id(self, tool_id):
        self.reloaded[tool_id] = True

    def was_reloaded(self, tool_id):
        return self.reloaded.get(tool_id, False)


class CallbackRecorder:
    def __init__(self):
        self.called = False

    def call(self):
        self.called = True


@contextmanager
def __test_directory():
    base_path = tempfile.mkdtemp()
    try:
        yield base_path
    finally:
        rmtree(base_path)



from unittest.mock import Mock

from galaxy.tool_util.output_checker import (
    check_output,
    DETECTED_JOB_STATE,
)
from galaxy.tool_util.parser.stdio import (
    StdioErrorLevel,
    ToolStdioRegex,
)
from galaxy.util.unittest import TestCase


class TestOutputChecker(TestCase):
    def setUp(self):
        self.tool = Mock(
            stdio_regexes=[],
            stdio_exit_codes=[],
        )
        self.stdout = ""
        self.stderr = ""
        self.tool_exit_code = 0

    def test_default_no_stderr_success(self):
        self.__assertSuccessful()

    def test_default_stderr_failure(self):
        self.stderr = "foo"
        self.__assertNotSuccessful()

    def test_exit_code_error(self):
        mock_exit_code = Mock(range_start=1, range_end=1, error_level=StdioErrorLevel.FATAL, desc=None)
        self.tool.stdio_exit_codes.append(mock_exit_code)
        self.tool_exit_code = 1
        self.__assertNotSuccessful()

    def test_exit_code_success(self):
        mock_exit_code = Mock(range_start=1, range_end=1, error_level=StdioErrorLevel.FATAL, desc=None)
        self.tool.stdio_exit_codes.append(mock_exit_code)
        self.tool_exit_code = 0
        self.__assertSuccessful()

    def test_problematic_strings_matching(self):
        problematic_str = "\x80abc"
        self.__add_regex(
            Mock(match=r".abc", stdout_match=False, stderr_match=True, error_level=StdioErrorLevel.FATAL, desc=None)
        )
        self.stderr = problematic_str
        self.__assertNotSuccessful()

    def test_problematic_strings_not_matching(self):
        problematic_str = "\x80abc"
        self.__add_regex(
            Mock(match=r".abcd", stdout_match=False, stderr_match=True, error_level=StdioErrorLevel.FATAL, desc=None)
        )
        self.stderr = problematic_str
        self.__assertSuccessful()

    def test_stderr_regex_negative_match(self):
        regex = ToolStdioRegex()
        regex.stderr_match = True
        regex.match = "foobar"
        self.__add_regex(regex)
        self.stderr = "foo"
        self.__assertSuccessful()

    def test_stderr_regex_positive_match(self):
        regex = ToolStdioRegex()
        regex.stderr_match = True
        regex.match = "foo"
        self.__add_regex(regex)
        self.stderr = "foobar"
        self.__assertNotSuccessful()

    def test_stderr_regex_group_positive_match(self):
        regex = ToolStdioRegex()
        regex.stderr_match = True
        regex.match = "ERROR: (.{3})bar"
        regex.desc = r"\1"
        self.__add_regex(regex)
        self.stderr = "ERROR: foobar"
        _, _, _, job_messages = self.__check_output()
        assert job_messages[0]["desc"] == "Fatal error: foo"
        self.__assertNotSuccessful()

    def test_stderr_regex_namedgroup_positive_match(self):
        regex = ToolStdioRegex()
        regex.stderr_match = True
        regex.match = "ERROR: (?P<q>.{3})bar"
        regex.desc = r"\g<q>"
        self.__add_regex(regex)
        self.stderr = "ERROR: foobar"
        _, _, _, job_messages = self.__check_output()
        assert job_messages[0]["desc"] == "Fatal error: foo"
        self.__assertNotSuccessful()

    def test_stdout_ignored_for_stderr_regexes(self):
        regex = ToolStdioRegex()
        regex.stderr_match = True
        regex.match = "foo"
        self.__add_regex(regex)
        self.stdout = "foobar"
        self.__assertSuccessful()

    def test_stderr_ignored_for_stdout_regexes(self):
        regex = ToolStdioRegex()
        regex.stdout_match = True
        regex.match = "foo"
        self.__add_regex(regex)
        self.stderr = "foobar"
        self.__assertSuccessful()

    def __add_regex(self, regex):
        self.tool.stdio_regexes.append(regex)

    def __assertSuccessful(self):
        assert self.__check_output()[0] == DETECTED_JOB_STATE.OK

    def __assertNotSuccessful(self):
        assert self.__check_output()[0] != DETECTED_JOB_STATE.OK

    def __check_output(self):
        return check_output(
            self.tool.stdio_regexes, self.tool.stdio_exit_codes, self.stdout, self.stderr, self.tool_exit_code
        )



from typing import Optional

from galaxy.tool_util.parser.parameter_validators import parse_xml_validators
from galaxy.tool_util.unittest_utils.sample_data import (
    INVALID_XML_VALIDATORS,
    VALID_XML_VALIDATORS,
)
from galaxy.util import XML


def test_xml_validation_valid():
    for xml_validator in VALID_XML_VALIDATORS:
        _validate_xml_str(xml_validator)


def test_xml_validation_invalid():
    for xml_validator in INVALID_XML_VALIDATORS:
        exc: Optional[Exception] = None
        try:
            _validate_xml_str(xml_validator)
        except ValueError as e:
            exc = e
        assert exc is not None, f"{xml_validator} - validated when it wasn't expected to"


def _validate_xml_str(xml_str: str):
    xml_el = XML(f"<param>{xml_str}</param>")
    parse_xml_validators(xml_el)



import os
import os.path
import shutil
import tempfile
from math import isinf
from typing import (
    Optional,
    Sequence,
    Type,
    TypeVar,
)

from galaxy.tool_util.parser.factory import get_tool_source
from galaxy.tool_util.parser.output_objects import from_tool_source
from galaxy.tool_util.unittest_utils import functional_test_tool_path
from galaxy.tool_util_models.tool_outputs import (
    ToolOutput,
    ToolOutputCollection,
    ToolOutputDataset,
)
from galaxy.util import galaxy_directory
from galaxy.util.resources import (
    as_file,
    resource_path,
)
from galaxy.util.unittest import TestCase

TOOL_XML_1 = """
<tool name="BWA Mapper" id="bwa" version="1.0.1" display_interface="true" require_login="true" hidden="true">
    <description>The BWA Mapper</description>
    <xrefs>
        <xref type="bio.tools">bwa</xref>
    </xrefs>
    <creator>
        <person
            givenName="Björn"
            familyName="Grüning"
            identifier="http://orcid.org/0000-0002-3079-6586" />
        <organization
            url="https://galaxyproject.org/iuc/"
            name="Galaxy IUC" />
    </creator>
    <version_command interpreter="python">bwa.py --version</version_command>
    <parallelism method="multi" split_inputs="input1" split_mode="to_size" split_size="1" merge_outputs="out_file1" />
    <command interpreter="python">bwa.py --arg1=42</command>
    <requirements>
        <container type="docker">mycool/bwa</container>
        <requirement type="package" version="1.0">bwa</requirement>
        <resource type="cores_min">1</resource>
        <resource type="cuda_version_min">10.2</resource>
        <resource type="cuda_compute_capability">6.1</resource>
        <resource type="gpu_memory_min">4042</resource>
        <resource type="cuda_device_count_min">1</resource>
        <resource type="cuda_device_count_max">2</resource>
        <resource type="shm_size">67108864</resource>
        <credentials name="Apollo" version="gmod.org/apollo" label="Apollo credential set" description="Please provide credentials for Apollo">
            <variable name="server" inject_as_env="apollo_url" optional="true" label="Your Apollo server" description="URL of your Apollo server" />
            <secret name="username" inject_as_env="apollo_user" optional="true" label="Your Apollo username" description="Username for Apollo" />
            <secret name="password" inject_as_env="apollo_pass" optional="true" label="Your Apollo password" description="Password for Apollo" />
        </credentials>
    </requirements>
    <outputs>
        <data name="out1" format="bam" from_work_dir="out1.bam" />
    </outputs>
    <stdio>
        <exit_code range="1:" level="fatal" />
    </stdio>
    <help>This is HELP TEXT1!!!</help>
    <tests>
        <test>
            <param name="foo" value="5" />
            <output name="out1" file="moo.txt" />
        </test>
        <test>
            <param name="foo" value="5">
            </param>
            <output name="out1" lines_diff="4" compare="sim_size">
                <metadata name="dbkey" value="hg19" />
            </output>
        </test>
    </tests>
</tool>
"""

TOOL_WITH_TOKEN = r"""
<tool id="tool_with_token" name="Token" version="1">
    <macros>
        <token name="@ESCAPE_IDENTIFIER@">
<![CDATA[
#set identifier = re.sub('[^\s\w\-]', '_', str($file.element_identifier))
        ]]></token>
        <token name="@NESTED_TOKEN@">
<![CDATA[
    before
    @ESCAPE_IDENTIFIER@
    after
        ]]></token>
    </macros>
    <command>
@NESTED_TOKEN@
    </command>
</tool>
"""

TOOL_WTIH_TOKEN_FROM_MACRO_FILE = r"""
<tool id="tool_with_token" name="Token" version="1">
    <macros>
        <import>macros.xml</import>
    </macros>
    <command detect_errors="exit_code"><![CDATA[@CATS@]]></command>
</tool>
"""

MACRO_CONTENTS = r"""<?xml version="1.0"?>
<macros>
    <token name="@CATS@">cat   </token>
</macros>
"""

TOOL_WITH_RECURSIVE_TOKEN = r"""
<tool id="tool_with_recursive_token" name="Token" version="1">
    <macros>
        <token name="@NESTED_TOKEN@">
<![CDATA[
    before
    @NESTED_TOKEN@
    after
        ]]></token>
    </macros>
    <command>
@NESTED_TOKEN@
    </command>
</tool>
"""

TOOL_YAML_1 = """
name: "Bowtie Mapper"
class: GalaxyTool
id: bowtie
version: 1.0.2
description: "The Bowtie Mapper"
xrefs:
  - type: bio.tools
    value: bwa
command: "bowtie_wrapper.pl --map-the-stuff"
interpreter: "perl"
runtime_version:
  command: "bowtie --version"
requirements:
  - type: package
    name: bwa
    version: 1.0.1
  - type: resource
    cores_min: 1
  - type: resource
    cuda_version_min: 10.2
  - type: resource
    cuda_compute_capability: 6.1
  - type: resource
    gpu_memory_min: 4042
  - type: resource
    cuda_device_count_min: 1
  - type: resource
    cuda_device_count_max: 2
  - type: resource
    shm_size: 67108864
containers:
  - type: docker
    identifier: "awesome/bowtie"
credentials:
  - name: Apollo
    version: gmod.org/apollo
    secrets:
      - name: username
        label: Your Apollo username
        description: Username for Apollo
        inject_as_env: apollo_user
        optional: true
      - name: password
        label: Your Apollo password
        description: Password for Apollo
        inject_as_env: apollo_pass
        optional: true
    variables:
      - name: server
        label: Your Apollo server
        description: URL of your Apollo server
        inject_as_env: apollo_url
        optional: true
outputs:
  out1:
    format: bam
    from_work_dir: out1.bam
inputs:
  - name: input1
    type: integer
    min: 7
    max: 8
  - name: moo
    label: cow
    type: repeat
    blocks:
      - name: nestinput
        type: data
      - name: nestsample
        type: text
help:
|
    This is HELP TEXT2!!!
tests:
   - inputs:
       foo: 5
     outputs:
       out1: moo.txt
   - inputs:
       foo:
         value: 5
     outputs:
       out1:
         lines_diff: 4
         compare: sim_size
"""

TOOL_EXPRESSION_XML_1 = """
<tool name="parse_int" id="parse_int" version="0.1.0" tool_type="expression">
    <description>Parse Int</description>
    <expression type="ecma5.1">
        {return {'output': parseInt($job.input1)};}
    </expression>
    <inputs>
        <input type="text" label="Text to parse." name="input1" />
    </inputs>
    <outputs>
        <output type="integer" name="out1" from="output" />
    </outputs>
    <help>Parse an integer from text.</help>
</tool>
"""


TOOL_EXPRESSION_YAML_1 = """
class: GalaxyExpressionTool
name: "parse_int"
id: parse_int
version: 1.0.2
expression: "{return {'output': parseInt($job.input1)};}"
inputs:
 - name: input1
   label: Text to parse
   type: text
outputs:
  out1:
    type: integer
    from: "#output"
"""


def get_test_tool_source(source_file_name=None, source_contents=None, macro_contents=None, temp_directory=None):
    source_directory = temp_directory or tempfile.mkdtemp()
    macro_paths = []
    if not os.path.isabs(source_file_name):
        path = os.path.join(source_directory, source_file_name)
        with open(path, "w") as out:
            out.write(source_contents)
        if macro_contents:
            macro_path = os.path.join(source_directory, "macros.xml")
            with open(macro_path, "w") as out:
                out.write(macro_contents)
            macro_paths = [macro_path]
    else:
        path = source_file_name
    tool_source = get_tool_source(path, macro_paths=macro_paths)
    if temp_directory is None:
        shutil.rmtree(source_directory)
    return tool_source


class BaseLoaderTestCase(TestCase):
    source_file_name: Optional[str] = None
    source_contents: Optional[str] = None

    def setUp(self):
        self.temp_directory = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_directory)

    @property
    def _tool_source(self):
        return self._get_tool_source()

    @property
    def _output_models(self) -> Sequence[ToolOutput]:
        return from_tool_source(self._tool_source)

    def _get_tool_source(self, source_file_name=None, source_contents=None, macro_contents=None):
        if source_file_name is None:
            source_file_name = self._get_source_file_name()
        if source_contents is None:
            source_contents = self.source_contents

        return get_test_tool_source(
            source_file_name,
            source_contents,
            macro_contents,
            self.temp_directory,
        )

    def _get_source_file_name(self) -> str:
        assert self.source_file_name
        return self.source_file_name


class TestXmlExpressionLoader(BaseLoaderTestCase):
    source_file_name = "expression.xml"
    source_contents = TOOL_EXPRESSION_XML_1

    def test_expression(self):
        assert self._tool_source.parse_expression().strip() == "{return {'output': parseInt($job.input1)};}"

    def test_tool_type(self):
        assert self._tool_source.parse_tool_type() == "expression"


class TestXmlLoader(BaseLoaderTestCase):
    source_file_name = "bwa.xml"
    source_contents = TOOL_XML_1

    def test_tool_source_to_string(self):
        # Previously this threw an Exception - test for regression.
        str(self._tool_source)

    def test_version(self):
        assert self._tool_source.parse_version() == "1.0.1"

    def test_id(self):
        assert self._tool_source.parse_id() == "bwa"

    def test_module_and_type(self):
        assert self._tool_source.parse_tool_module() is None
        assert self._tool_source.parse_tool_type() is None

    def test_name(self):
        assert self._tool_source.parse_name() == "BWA Mapper"

    def test_display_interface(self):
        assert self._tool_source.parse_display_interface(False)

    def test_require_login(self):
        assert self._tool_source.parse_require_login(False)

    def test_parse_request_param_translation_elem(self):
        assert self._tool_source.parse_request_param_translation_elem() is None

    def test_command_parsing(self):
        assert self._tool_source.parse_command() == "bwa.py --arg1=42"
        assert self._tool_source.parse_interpreter() == "python"

    def test_descripting_parsing(self):
        assert self._tool_source.parse_description() == "The BWA Mapper"

    def test_version_command(self):
        assert self._tool_source.parse_version_command() == "bwa.py --version"
        assert self._tool_source.parse_version_command_interpreter() == "python"

    def test_parallelism(self):
        parallelism_info = self._tool_source.parse_parallelism()
        assert parallelism_info.method == "multi"
        assert parallelism_info.attributes["split_inputs"] == "input1"

    def test_hidden(self):
        assert self._tool_source.parse_hidden()

    def test_action(self):
        assert self._tool_source.parse_action_module() is None

    def test_requirements(self):
        requirements, containers, resource_requirements, _, credentials = self._tool_source.parse_requirements()
        assert requirements[0].type == "package"
        assert list(containers)[0].identifier == "mycool/bwa"
        assert resource_requirements[0].resource_type == "cores_min"
        assert resource_requirements[1].resource_type == "cuda_version_min"
        assert resource_requirements[2].resource_type == "cuda_compute_capability"
        assert resource_requirements[3].resource_type == "gpu_memory_min"
        assert resource_requirements[4].resource_type == "cuda_device_count_min"
        assert resource_requirements[5].resource_type == "cuda_device_count_max"
        assert resource_requirements[6].resource_type == "shm_size"
        assert not resource_requirements[0].runtime_required
        assert credentials[0].name == "Apollo"
        assert credentials[0].version == "gmod.org/apollo"
        assert len(credentials[0].secrets) == 2
        assert len(credentials[0].variables) == 1

    def test_outputs(self):
        outputs, output_collections = self._tool_source.parse_outputs(object())
        assert len(outputs) == 1, outputs
        assert len(output_collections) == 0

    def test_stdio(self):
        exit, regexes = self._tool_source.parse_stdio()
        assert len(exit) == 1
        assert len(regexes) == 0
        assert exit[0].range_start == 1
        assert isinf(exit[0].range_end)

    def test_help(self):
        help_text = self._tool_source.parse_help().content
        assert help_text.strip() == "This is HELP TEXT1!!!"

    def test_tests(self):
        tests_dict = self._tool_source.parse_tests_to_dict()
        tests = tests_dict["tests"]
        assert len(tests) == 2
        test_dict = tests[0]
        inputs = test_dict["inputs"]
        assert len(inputs) == 1, test_dict
        input1 = inputs[0]
        assert input1["name"] == "foo", input1
        assert input1["value"] == "5"

        outputs = test_dict["outputs"]
        assert len(outputs) == 1
        output1 = outputs[0]
        assert output1["name"] == "out1"
        assert output1["value"] == "moo.txt"
        attributes1 = output1["attributes"]
        assert attributes1["compare"] == "diff"
        assert attributes1["lines_diff"] == 0

        test2 = tests[1]
        outputs = test2["outputs"]
        assert len(outputs) == 1
        output2 = outputs[0]
        assert output2["name"] == "out1"
        assert output2["value"] is None
        attributes1 = output2["attributes"]
        assert attributes1["compare"] == "sim_size"
        assert attributes1["lines_diff"] == 4

    def test_output_models(self):
        output_models = self._output_models
        assert len(output_models) == 1
        output_model = output_models[0]
        assert output_model.name == "out1"
        assert not output_model.hidden
        assert output_model.label is None
        output_dataset_model = assert_output_model_of_type(output_model, ToolOutputDataset)
        assert output_dataset_model.metadata_source is None

    def test_xrefs(self):
        xrefs = self._tool_source.parse_xrefs()
        assert xrefs == [{"value": "bwa", "type": "bio.tools"}]

    def test_exit_code(self):
        tool_source = self._get_tool_source(source_contents="""<tool id="bwa" name="bwa">
            <command detect_errors="exit_code">
                ls
            </command>
        </tool>
        """)
        exit, regexes = tool_source.parse_stdio()
        assert len(exit) == 2, exit
        assert len(regexes) == 0, regexes

        tool_source = self._get_tool_source(source_contents="""<tool id="bwa" name="bwa">
            <command detect_errors="aggressive">
                ls
            </command>
        </tool>
        """)
        exit, regexes = tool_source.parse_stdio()
        assert len(exit) == 2, exit
        # error:, exception: various memory exception...
        assert len(regexes) > 2, regexes

    def test_sanitize_option(self):
        assert self._tool_source.parse_sanitize() is True

    def test_refresh_option(self):
        assert self._tool_source.parse_refresh() is False

    def test_nested_token(self):
        tool_source = self._get_tool_source(source_contents=TOOL_WITH_TOKEN)
        command = tool_source.parse_command()
        assert command
        assert "@" not in command

    def test_token_significant_whitespace(self):
        tool_source = self._get_tool_source(
            source_contents=TOOL_WTIH_TOKEN_FROM_MACRO_FILE, macro_contents=MACRO_CONTENTS
        )
        command = tool_source.parse_command()
        assert command == "cat   "
        assert "@" not in command

    def test_recursive_token(self):
        with self.assertRaisesRegex(Exception, "^Token '@NESTED_TOKEN@' cannot contain itself$"):
            self._get_tool_source(source_contents=TOOL_WITH_RECURSIVE_TOKEN)

    def test_creator(self):
        creators = self._tool_source.parse_creator()
        assert len(creators) == 2
        creator1 = creators[0]
        assert creator1["class"] == "Person"
        assert creator1["identifier"] == "http://orcid.org/0000-0002-3079-6586"

        creator2 = creators[1]
        assert creator2["class"] == "Organization"
        assert creator2["name"] == "Galaxy IUC"


class TestYamlLoader(BaseLoaderTestCase):
    source_file_name = "bwa.yml"
    source_contents = TOOL_YAML_1

    def test_version(self):
        assert self._tool_source.parse_version() == "1.0.2"

    def test_id(self):
        assert self._tool_source.parse_id() == "bowtie"

    def test_module_and_type(self):
        # These just rely on defaults
        assert self._tool_source.parse_tool_module() is None
        assert self._tool_source.parse_tool_type() is None

    def test_name(self):
        assert self._tool_source.parse_name() == "Bowtie Mapper"

    def test_display_interface(self):
        assert not self._tool_source.parse_display_interface(False)
        assert self._tool_source.parse_display_interface(True)

    def test_require_login(self):
        assert not self._tool_source.parse_require_login(False)

    def test_parse_request_param_translation_elem(self):
        assert self._tool_source.parse_request_param_translation_elem() is None

    def test_command_parsing(self):
        assert self._tool_source.parse_command() == "bowtie_wrapper.pl --map-the-stuff"
        assert self._tool_source.parse_interpreter() == "perl"

    def test_parse_redirect_url_params_elem(self):
        assert self._tool_source.parse_redirect_url_params_elem() is None

    def test_descripting_parsing(self):
        assert self._tool_source.parse_description() == "The Bowtie Mapper"

    def test_version_command(self):
        assert self._tool_source.parse_version_command() == "bowtie --version"
        assert self._tool_source.parse_version_command_interpreter() is None

    def test_parallelism(self):
        assert self._tool_source.parse_parallelism() is None

    def test_hidden(self):
        assert not self._tool_source.parse_hidden()

    def test_action(self):
        assert self._tool_source.parse_action_module() is None

    def test_requirements(self):
        software_requirements, containers, resource_requirements, _, credentials = (
            self._tool_source.parse_requirements()
        )
        assert software_requirements.to_dict() == [{"name": "bwa", "type": "package", "version": "1.0.1", "specs": []}]
        assert len(containers) == 1
        assert containers[0].to_dict() == {
            "identifier": "awesome/bowtie",
            "type": "docker",
            "resolve_dependencies": False,
            "shell": "/bin/sh",
        }
        assert len(resource_requirements) == 7
        assert resource_requirements[0].to_dict() == {"resource_type": "cores_min", "value_or_expression": 1}
        assert resource_requirements[1].to_dict() == {"resource_type": "cuda_version_min", "value_or_expression": 10.2}
        assert resource_requirements[2].to_dict() == {
            "resource_type": "cuda_compute_capability",
            "value_or_expression": 6.1,
        }
        assert resource_requirements[3].to_dict() == {"resource_type": "gpu_memory_min", "value_or_expression": 4042}
        assert resource_requirements[4].to_dict() == {
            "resource_type": "cuda_device_count_min",
            "value_or_expression": 1,
        }
        assert resource_requirements[5].to_dict() == {
            "resource_type": "cuda_device_count_max",
            "value_or_expression": 2,
        }
        assert resource_requirements[6].to_dict() == {
            "resource_type": "shm_size",
            "value_or_expression": 67108864,
        }
        assert len(credentials) == 1
        assert len(credentials[0].secrets) == 2
        assert len(credentials[0].variables) == 1

    def test_outputs(self):
        outputs, output_collections = self._tool_source.parse_outputs(object())
        assert len(outputs) == 1
        assert len(output_collections) == 0

    def test_stdio(self):
        exit, regexes = self._tool_source.parse_stdio()
        assert len(exit) == 2

        assert isinf(exit[0].range_start)
        assert exit[0].range_start == float("-inf")

        assert exit[1].range_start == 1
        assert isinf(exit[1].range_end)

    def test_help(self):
        help_text = self._tool_source.parse_help().content
        assert help_text.strip() == "This is HELP TEXT2!!!"

    def test_inputs(self):
        input_pages = self._tool_source.parse_input_pages()
        assert input_pages.inputs_defined
        page_sources = input_pages.page_sources
        assert len(page_sources) == 1
        page_source = page_sources[0]
        input_sources = page_source.parse_input_sources()
        assert len(input_sources) == 2

    def test_tests(self):
        tests_dict = self._tool_source.parse_tests_to_dict()
        tests = tests_dict["tests"]
        assert len(tests) == 2
        test_dict = tests[0]
        inputs = test_dict["inputs"]
        assert len(inputs) == 1
        input1 = inputs[0]
        assert input1["name"] == "foo"
        assert input1["value"] == 5

        outputs = test_dict["outputs"]
        assert len(outputs) == 1
        output1 = outputs[0]
        assert output1["name"] == "out1"
        assert output1["value"] == "moo.txt"
        attributes1 = output1["attributes"]
        assert attributes1["compare"] == "diff"
        assert attributes1["lines_diff"] == 0

        test2 = tests[1]
        outputs = test2["outputs"]
        assert len(outputs) == 1
        output2 = outputs[0]
        assert output2["name"] == "out1"
        assert output2["value"] is None
        attributes1 = output2["attributes"]
        assert attributes1["compare"] == "sim_size"
        assert attributes1["lines_diff"] == 4

    def test_xrefs(self):
        xrefs = self._tool_source.parse_xrefs()
        assert xrefs == [{"value": "bwa", "type": "bio.tools"}]

    def test_sanitize(self):
        assert self._tool_source.parse_sanitize() is True


class TestDataSourceLoader(BaseLoaderTestCase):
    source_file_name = "ds.xml"
    source_contents = """<?xml version="1.0"?>
<tool name="YeastMine" id="yeastmine" tool_type="data_source">
    <description>server</description>
    <command interpreter="python">data_source.py $output $__app__.config.output_size_limit</command>
    <inputs action="http://yeastmine.yeastgenome.org/yeastmine/begin.do" check_values="false" method="get">
        <display>go to yeastMine server $GALAXY_URL</display>
    </inputs>
    <request_param_translation>
        <request_param galaxy_name="data_type" remote_name="data_type" missing="auto" >
            <value_translation>
                <value galaxy_value="auto" remote_value="txt" /> <!-- intermine currently always provides 'txt', make this auto detect -->
            </value_translation>
        </request_param>
    </request_param_translation>
    <!-- Following block doesn't really belong here - not sure what block is suppose to do actually cannot
         find actual usage. -->
    <redirect_url_params>cow</redirect_url_params>
    <uihints minwidth="800"/>
    <outputs>
        <data name="output" format="txt" />
    </outputs>
    <options sanitize="False" refresh="True"/>
</tool>
"""

    def test_sanitize_option(self):
        assert self._tool_source.parse_sanitize() is False

    def test_refresh_option(self):
        assert self._tool_source.parse_refresh() is True

    def test_tool_type(self):
        assert self._tool_source.parse_tool_type() == "data_source"

    def test_parse_request_param_translation_elem(self):
        assert self._tool_source.parse_request_param_translation_elem() is not None

    def test_redirect_url_params_elem(self):
        assert self._tool_source.parse_redirect_url_params_elem() is not None

    def test_parallelism(self):
        assert self._tool_source.parse_parallelism() is None

    def test_hidden(self):
        assert not self._tool_source.parse_hidden()


class TestApplyRulesToolLoader(BaseLoaderTestCase):
    source_file_name = os.path.join(galaxy_directory(), "lib/galaxy/tools/apply_rules.xml")
    source_contents = None

    def test_tool_type(self):
        tool_module = self._tool_source.parse_tool_module()
        assert tool_module[0] == "galaxy.tools"
        assert tool_module[1] == "ApplyRulesTool"
        assert self._tool_source.parse_tool_type() == "apply_rules_to_collection"

    def test_outputs(self):
        outputs, output_collections = self._tool_source.parse_outputs(object())
        assert len(outputs) == 1
        assert len(output_collections) == 1

    def test_output_models(self):
        output_models = self._output_models
        assert len(output_models) == 1
        output_model = output_models[0]
        assert output_model.name == "output"
        assert not output_model.hidden
        assert output_model.label == "${input.name} (re-organized)"
        output_collection_model = assert_output_model_of_type(output_model, ToolOutputCollection)
        structure = output_collection_model.structure
        assert structure.collection_type_from_rules == "rules"


class TestBuildListToolLoader(BaseLoaderTestCase):
    source_file_name = os.path.join(galaxy_directory(), "lib/galaxy/tools/build_list.xml")
    source_contents = None

    def test_tool_type(self):
        tool_module = self._tool_source.parse_tool_module()
        assert tool_module[0] == "galaxy.tools"
        assert tool_module[1] == "BuildListCollectionTool"


class FunctionalTestToolTestCase(BaseLoaderTestCase):
    test_path: str
    source_contents: None

    def _get_source_file_name(self) -> str:
        return functional_test_tool_path(self.test_path)


class TestExpressionTestToolLoader(FunctionalTestToolTestCase):
    test_path = "expression_null_handling_boolean.xml"

    def test_test(self):
        test_dicts = self._tool_source.parse_tests_to_dict()["tests"]
        assert len(test_dicts) == 3
        test_dict_0 = test_dicts[0]
        assert "outputs" in test_dict_0, test_dict_0
        outputs = test_dict_0["outputs"]
        output0 = outputs[0]
        assert "object" in output0["attributes"]
        assert output0["attributes"]["object"] is True

        test_dict_1 = test_dicts[1]
        assert "outputs" in test_dict_1, test_dict_1
        outputs = test_dict_1["outputs"]
        output0 = outputs[0]
        assert "object" in output0["attributes"]
        assert output0["attributes"]["object"] is False

        test_dict_2 = test_dicts[2]
        assert "outputs" in test_dict_2, test_dict_2
        outputs = test_dict_2["outputs"]
        output0 = outputs[0]
        assert "object" in output0["attributes"]
        assert output0["attributes"]["object"] is None

    def test_output_models(self):
        output_models = self._output_models
        assert len(output_models) == 1
        output_model = output_models[0]
        assert output_model.name == "bool_out"
        assert not output_model.hidden
        assert output_model.label is None


class TestDefaultDataTestToolLoader(FunctionalTestToolTestCase):
    test_path = "for_workflows/cat_default.xml"

    def test_input_parsing(self):
        input_pages = self._tool_source.parse_input_pages()
        assert input_pages.inputs_defined
        page_sources = input_pages.page_sources
        assert len(page_sources) == 1
        page_source = page_sources[0]
        input_sources = page_source.parse_input_sources()
        assert len(input_sources) == 1
        data_input = input_sources[0]
        default_dict = data_input.parse_default()
        assert default_dict
        assert default_dict["location"] == "https://raw.githubusercontent.com/galaxyproject/galaxy/dev/test-data/1.bed"


class TestDefaultCollectionDataTestToolLoader(FunctionalTestToolTestCase):
    test_path = "collection_paired_default.xml"

    def test_input_parsing(self):
        input_pages = self._tool_source.parse_input_pages()
        assert input_pages.inputs_defined
        page_sources = input_pages.page_sources
        assert len(page_sources) == 1
        page_source = page_sources[0]
        input_sources = page_source.parse_input_sources()
        assert len(input_sources) == 1
        data_input = input_sources[0]
        default_dict = data_input.parse_default()
        assert default_dict
        assert default_dict["collection_type"] == "paired"
        elements = default_dict["elements"]
        assert len(elements) == 2
        element0 = elements[0]
        assert element0["identifier"] == "forward"
        assert element0["location"] == "https://raw.githubusercontent.com/galaxyproject/galaxy/dev/test-data/1.bed"
        element1 = elements[1]
        assert element1["identifier"] == "reverse"
        assert element1["location"] == "https://raw.githubusercontent.com/galaxyproject/galaxy/dev/test-data/1.fasta"


class TestDefaultNestedCollectionDataTestToolLoader(FunctionalTestToolTestCase):
    test_path = "collection_nested_default.xml"

    def test_input_parsing(self):
        input_pages = self._tool_source.parse_input_pages()
        assert input_pages.inputs_defined
        page_sources = input_pages.page_sources
        assert len(page_sources) == 1
        page_source = page_sources[0]
        input_sources = page_source.parse_input_sources()
        assert len(input_sources) == 1
        data_input = input_sources[0]
        default_dict = data_input.parse_default()
        assert default_dict
        assert default_dict["collection_type"] == "list:paired"
        elements = default_dict["elements"]
        assert len(elements) == 1
        element0 = elements[0]
        assert element0["identifier"] == "i1"

        elements0 = element0["elements"]
        assert len(elements0) == 2
        elements00 = elements0[0]
        assert elements00["identifier"] == "forward"
        elements01 = elements0[1]
        assert elements01["identifier"] == "reverse"


class TestExpressionOutputDataToolLoader(FunctionalTestToolTestCase):
    test_path = "expression_pick_larger_file.xml"

    def test_output_parsing(self):
        outputs, _ = self._tool_source.parse_outputs(None)
        assert "larger_file" in outputs
        tool_output = outputs["larger_file"]
        assert tool_output.format == "data"
        assert tool_output.from_expression == "output"


class TestSpecialToolLoader(BaseLoaderTestCase):
    source_file_name = os.path.join(galaxy_directory(), "lib/galaxy/tools/imp_exp/exp_history_to_archive.xml")
    source_contents = None

    def test_tool_type(self):
        tool_module = self._tool_source.parse_tool_module()
        # Probably we don't parse_tool_module any more? -
        # tool_type seems sufficient.
        assert tool_module[0] == "galaxy.tools"
        assert tool_module[1] == "ExportHistoryTool"
        assert self._tool_source.parse_tool_type() == "export_history"

    def test_version_command(self):
        assert self._tool_source.parse_version_command() is None
        assert self._tool_source.parse_version_command_interpreter() is None

    def test_action(self):
        action = self._tool_source.parse_action_module()
        assert action[0] == "galaxy.tools.actions.history_imp_exp"
        assert action[1] == "ExportHistoryToolAction"


class TestCollection(FunctionalTestToolTestCase):
    test_path = "collection_two_paired.xml"

    def test_tests(self):
        tests_dict = self._tool_source.parse_tests_to_dict()
        tests = tests_dict["tests"]
        assert len(tests) == 2
        assert len(tests[0]["inputs"]) == 3, tests[0]

        outputs, output_collections = self._tool_source.parse_outputs(None)
        assert len(output_collections) == 0


class TestCollectionOutputXml(FunctionalTestToolTestCase):
    test_path = "collection_creates_pair.xml"

    def test_tests(self):
        outputs, output_collections = self._tool_source.parse_outputs(None)
        assert len(output_collections) == 1


class TestCollectionOutputYaml(FunctionalTestToolTestCase):
    test_path = "collection_creates_pair_y.yml"

    def test_tests(self):
        outputs, output_collections = self._tool_source.parse_outputs(None)
        assert len(output_collections) == 1


class TestEnvironmentVariables(FunctionalTestToolTestCase):
    test_path = "environment_variables.xml"

    def test_tests(self):
        tests_dict = self._tool_source.parse_tests_to_dict()
        tests = tests_dict["tests"]
        assert len(tests) == 1


class TestExpectations(FunctionalTestToolTestCase):
    test_path = "detect_errors.xml"

    def test_tests(self):
        tests_dict = self._tool_source.parse_tests_to_dict()
        tests = tests_dict["tests"]
        assert len(tests) == 10
        test_0 = tests[0]
        assert len(test_0["stderr"]) == 1
        assert len(test_0["stdout"]) == 2


class TestExpectationsCommandVersion(FunctionalTestToolTestCase):
    test_path = "job_properties.xml"

    def test_tests(self):
        tests_dict = self._tool_source.parse_tests_to_dict()
        tests = tests_dict["tests"]
        assert len(tests) > 0
        test_0 = tests[0]
        assert len(test_0["command_version"]) == 1


class TestQcStdio(FunctionalTestToolTestCase):
    test_path = "qc_stdout.xml"

    def test_tests(self):
        exit, regexes = self._tool_source.parse_stdio()
        assert len(exit) == 2
        assert len(regexes) == 2
        regex = regexes[0]
        assert regex.error_level == 1.1


class TestCollectionCatGroupTag(FunctionalTestToolTestCase):
    test_path = "collection_cat_group_tag.xml"

    def test_output_models(self):
        output_models = self._output_models
        assert len(output_models) == 1
        output_model = output_models[0]
        assert output_model.name == "out_file1"
        assert not output_model.hidden
        assert output_model.label is None
        output_dataset_model = assert_output_model_of_type(output_model, ToolOutputDataset)
        assert output_dataset_model.metadata_source == "input1"


def test_old_invalid_citation_dont_cause_failure_to_load():
    with as_file(resource_path(__name__, "invalid_citation.xml")) as tool_path:
        tool_source = get_tool_source(tool_path)
    assert tool_source.parse_citations() == []


def test_invalid_citation_not_allowed_in_modern_tools():
    with as_file(resource_path(__name__, "invalid_citation_24.2.xml")) as tool_path:
        tool_source = get_tool_source(tool_path)
    exc = None
    try:
        tool_source.parse_citations()
    except Exception as e:
        exc = e
    assert exc is not None


class TestToolProvidedMetadata2(FunctionalTestToolTestCase):
    test_path = "tool_provided_metadata_2.xml"

    def test_output_models(self):
        output_models = self._output_models
        assert len(output_models) == 1
        output_model = output_models[0]
        assert output_model.name == "sample"
        assert not output_model.hidden
        assert output_model.label is None
        output_dataset_model = assert_output_model_of_type(output_model, ToolOutputDataset)
        assert output_dataset_model.metadata_source is None
        discover_datasets = output_dataset_model.discover_datasets or []
        assert len(discover_datasets) == 1
        discover_datasets_0 = discover_datasets[0]
        assert discover_datasets_0.discover_via == "pattern"


T = TypeVar("T")


def assert_output_model_of_type(obj, clazz: Type[T]) -> T:
    assert isinstance(obj, clazz)
    return obj



import gzip
import os
import shutil
import tempfile
from typing import Tuple

try:
    import h5py
except ImportError:
    h5py = None
import pytest

from galaxy.tool_util.parser.xml import __parse_assert_list_from_elem
from galaxy.tool_util.verify import asserts
from galaxy.util import parse_xml_string

TABULAR_ASSERTION = """
    <assert_contents>
        <has_n_columns n="3"/>
    </assert_contents>
"""
TABULAR_CSV_ASSERTION = """
    <assert_contents>
        <has_n_columns sep="," min="3"/>
    </assert_contents>
"""
TABULAR_ASSERTION_COMMENT = """
    <assert_contents>
        <has_n_columns n="3" comment="#$"/>
    </assert_contents>
"""

TABULAR_DATA_POS = """1\t2\t3
"""

TABULAR_DATA_NEG = """1\t2\t3\t4
"""

TABULAR_CSV_DATA = """1,2
"""

TABULAR_DATA_COMMENT = """# comment
$ more comment (using a char with meaning wrt regexp)
1\t2\t3
"""


def test_has_n_columns_success():
    """test successful assertion"""
    a = run_assertions(TABULAR_ASSERTION, TABULAR_DATA_POS)
    assert len(a) == 0


def test_has_n_columns_failure():
    """test wrong number of columns"""
    a = run_assertions(TABULAR_ASSERTION, TABULAR_DATA_NEG)
    assert "Expected 3+-0 columns in output found 4" in a
    assert len(a) == 1


def test_has_n_columns_for_csv():
    """test wrong number of columns for csv data"""
    a = run_assertions(TABULAR_CSV_ASSERTION, TABULAR_CSV_DATA)
    assert "Expected the number of columns in output to be in [3:inf] found 2" in a
    assert len(a) == 1


def test_has_n_columns_with_comments():
    """test tabular data with comments"""
    a = run_assertions(TABULAR_ASSERTION_COMMENT, TABULAR_DATA_COMMENT)
    assert len(a) == 0


TEXT_DATA_HAS_TEXT = """test text
"""

TEXT_DATA_HAS_TEXT_NEG = """desired content
is not here
"""

TEXT_DATA_NONE = None

TEXT_DATA_EMPTY = ""

TEXT_HAS_TEXT_ASSERTION = """
    <assert_contents>
        <has_text text="test text"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N = """
    <assert_contents>
        <has_text text="test text" n="2"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N_DELTA = """
    <assert_contents>
        <has_text text="test text" n="3" delta="1"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_MIN_MAX = """
    <assert_contents>
        <has_text text="test text" min="2" max="4"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_NEGATE = """
    <assert_contents>
        <has_text text="test text" negate="true"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N_NEGATE = """
    <assert_contents>
        <has_text text="test text" n="2" negate="true"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N_DELTA_NEGATE = """
    <assert_contents>
        <has_text text="test text" n="3" delta="1" negate="true"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_MIN_MAX_NEGATE = """
    <assert_contents>
        <has_text text="test text" min="2" max="4" negate="true"/>
    </assert_contents>
"""

TEXT_NOT_HAS_TEXT_ASSERTION = """
    <assert_contents>
        <not_has_text text="not here"/>
    </assert_contents>
"""


def test_has_text_success():
    """test has_text"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_has_text_failure():
    """test has_text .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT_NEG)
    assert "Expected text 'test text' in output ('desired content\nis not here\n')" in a
    assert len(a) == 1


def test_has_text_None_output():
    """test has_text with None output"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_NONE)
    assert "Checking has_text assertion on empty output (None)" in a
    assert len(a) == 1


def test_has_text_empty_output():
    """test has_text with empty output"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_EMPTY)
    assert "Expected text 'test text' in output ('')" in a
    assert len(a) == 1


def test_has_text_n_success():
    """test has_text with n"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_text_n_failure():
    """test has_text with n .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N, TEXT_DATA_HAS_TEXT)
    assert "Expected 2+-0 occurences of 'test text' in output ('test text\n') found 1" in a
    assert len(a) == 1


def test_has_text_n_delta_success():
    """test has_text with n and delta"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N_DELTA, TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_text_n_delta_failure():
    """test has_text with n and delta .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N_DELTA, TEXT_DATA_HAS_TEXT)
    assert "Expected 3+-1 occurences of 'test text' in output ('test text\n') found 1" in a
    assert len(a) == 1


def test_has_text_minmax_delta_success():
    """test has_text with min max"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_MIN_MAX, TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_text_minmax_delta_failure():
    """test has_text with min max .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_MIN_MAX, TEXT_DATA_HAS_TEXT)
    assert "Expected that the number of occurences of 'test text' in output is in [2:4] ('test text\n') found 1" in a
    assert len(a) == 1


def test_has_text_negate_success():
    """test has_text negate"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_HAS_TEXT_NEG)
    assert len(a) == 0


def test_has_text_negate_failure():
    """test has_text negate .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_HAS_TEXT)
    assert "Did not expect text 'test text' in output ('test text\n')" in a
    assert len(a) == 1


def test_has_text_negate_None_output():
    """test has_text negate with None output .. should have the same output as with negate='false'"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_NONE)
    assert "Checking has_text assertion on empty output (None)" in a
    assert len(a) == 1


def test_has_text_negate_empty_output():
    """test has_text negate with empty output"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_EMPTY)
    assert len(a) == 0


def test_has_text_negate_n_success():
    """test has_text negate with n"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N_NEGATE, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_has_text_negate_n_failure():
    """test has_text negate with n .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N_NEGATE, TEXT_DATA_HAS_TEXT * 2)
    assert "Did not expect 2+-0 occurences of 'test text' in output ('test text\ntest text\n') found 2" in a
    assert len(a) == 1


def test_has_text_negate_n_delta_success():
    """test has_text negate with n and delta"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N_DELTA_NEGATE, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_has_text_negate_n_delta_failure():
    """test has_text negate with n and delta .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_N_DELTA_NEGATE, TEXT_DATA_HAS_TEXT * 2)
    assert "Did not expect 3+-1 occurences of 'test text' in output ('test text\ntest text\n') found 2" in a
    assert len(a) == 1


def test_has_text_negate_minmax_delta_success():
    """test has_text negate with min max"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_MIN_MAX_NEGATE, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_has_text_negate_minmax_delta_failure():
    """test has_text negate with min max .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_ASSERTION_MIN_MAX_NEGATE, TEXT_DATA_HAS_TEXT * 2)
    assert (
        "Did not expect that the number of occurences of 'test text' in output is in [2:4] ('test text\ntest text\n') found 2"
        in a
    )
    assert len(a) == 1


def test_not_has_text_success():
    """test not_has_text"""
    a = run_assertions(TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_not_has_text_failure():
    """test not_has_text .. negative test"""
    a = run_assertions(TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT_NEG)
    assert "Output file contains unexpected text 'not here'" in a
    assert len(a) == 1


def test_not_has_text_None_output():
    """test not_has_text with None output"""
    a = run_assertions(TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_NONE)
    assert "Checking not_has_text assertion on empty output (None)" in a
    assert len(a) == 1


def test_not_has_text_empty_output():
    """test not_has_text with empty output"""
    a = run_assertions(TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_EMPTY)
    assert len(a) == 0


TEXT_HAS_TEXT_MATCHING_ASSERTION = """
    <assert_contents>
        <has_text_matching expression="te[sx]t"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_MATCHING_ASSERTION_N = """
    <assert_contents>
        <has_text_matching expression="te[sx]t" n="4"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_MATCHING_ASSERTION_MINMAX = """
    <assert_contents>
        <has_text_matching expression="te[sx]t" min="3" max="5"/>
    </assert_contents>
"""


def test_has_text_matching_success():
    """test has_text_matching"""
    a = run_assertions(TEXT_HAS_TEXT_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_has_text_matching_failure():
    """test has_text_matching .. negative test"""
    a = run_assertions(TEXT_HAS_TEXT_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT_NEG)
    assert "Expected text matching expression 'te[sx]t' in output ('desired content\nis not here\n')" in a
    assert len(a) == 1


def test_has_text_matching_n_success():
    """test has_text_matching with n"""
    a = run_assertions(TEXT_HAS_TEXT_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_text_matching_n_failure():
    """test has_text_matching with n .. negative test (using the test text where "te[sx]st" appears twice)"""
    a = run_assertions(TEXT_HAS_TEXT_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT)
    assert "Expected 4+-0 (non-overlapping) matches for 'te[sx]t' in output ('test text\n') found 2" in a
    assert len(a) == 1


def test_has_text_matching_minmax_success():
    """test has_text_matching with min/max"""
    a = run_assertions(TEXT_HAS_TEXT_MATCHING_ASSERTION_MINMAX, TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_text_matching_minmax_failure():
    """test has_text_matching with n .. negative test (using the test text where "te[sx]st" appears twice)"""
    a = run_assertions(TEXT_HAS_TEXT_MATCHING_ASSERTION_MINMAX, TEXT_DATA_HAS_TEXT)
    assert (
        "Expected that the number of (non-overlapping) matches for 'te[sx]t' in output is in [3:5] ('test text\n') found 2"
        in a
    )
    assert len(a) == 1


TEXT_HAS_LINE_ASSERTION = """
    <assert_contents>
        <has_line line="test text"/>
    </assert_contents>
"""
TEXT_HAS_LINE_ASSERTION_N = """
    <assert_contents>
        <has_line line="test text" n="2"/>
    </assert_contents>
"""
TEXT_HAS_N_LINES_ASSERTION = """
    <assert_contents>
        <has_n_lines n="{n}"/>
    </assert_contents>
"""
TEXT_HAS_N_LINES_ASSERTION_DELTA = """
    <assert_contents>
        <has_n_lines n="{n}" delta="{delta}"/>
    </assert_contents>
"""
TEXT_HAS_LINE_MATCHING_ASSERTION = """
    <assert_contents>
        <has_line_matching expression="te[sx]t te[sx]t"/>
    </assert_contents>
"""
TEXT_HAS_LINE_MATCHING_ASSERTION_N = """
    <assert_contents>
        <has_line_matching expression="te[sx]t te[sx]t" n="2"/>
    </assert_contents>
"""


def test_has_line_success():
    """test has_line"""
    a = run_assertions(TEXT_HAS_LINE_ASSERTION, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_has_line_failure():
    """test has_line .. negative test"""
    a = run_assertions(TEXT_HAS_LINE_ASSERTION, TEXT_DATA_HAS_TEXT_NEG)
    assert "Expected line 'test text' in output ('desired content\nis not here\n')" in a
    assert len(a) == 1


def test_has_line_n_success():
    """test has_line with n"""
    a = run_assertions(TEXT_HAS_LINE_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_line_n_failure():
    """test has_line with n .. negative test"""
    a = run_assertions(TEXT_HAS_LINE_ASSERTION_N, TEXT_DATA_HAS_TEXT)
    assert "Expected 2+-0 lines 'test text' in output ('test text\n') found 1" in a
    assert len(a) == 1


def test_has_n_lines_success():
    """test has_n_lines"""
    a = run_assertions(TEXT_HAS_N_LINES_ASSERTION.format(n="2"), TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_n_lines_n_as_bytes_success():
    """test has_n_lines .. bytes"""
    a = run_assertions(TEXT_HAS_N_LINES_ASSERTION.format(n="2ki"), TEXT_DATA_HAS_TEXT * 2048)
    assert len(a) == 0


def test_has_n_lines_failure():
    """test has_n_lines .. negative test"""
    a = run_assertions(
        TEXT_HAS_N_LINES_ASSERTION.format(n="2"),
        TEXT_DATA_HAS_TEXT,
    )
    assert "Expected 2+-0 lines in the output found 1" in a
    assert len(a) == 1


def test_has_n_lines_delta():
    """test has_n_lines ..delta"""
    a = run_assertions(TEXT_HAS_N_LINES_ASSERTION_DELTA.format(n="3", delta="1"), TEXT_DATA_HAS_TEXT)
    assert "Expected 3+-1 lines in the output found 1" in a
    assert len(a) == 1


def test_has_line_matching_success():
    """test has_line_matching"""
    a = run_assertions(TEXT_HAS_LINE_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


def test_has_line_matching_failure():
    """test has_line_matching .. negative test"""
    a = run_assertions(TEXT_HAS_LINE_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT_NEG)
    assert "Expected line matching expression 'te[sx]t te[sx]t' in output ('desired content\nis not here\n')" in a
    assert len(a) == 1


def test_has_line_matching_n_success():
    """test has_line_matching n"""
    a = run_assertions(TEXT_HAS_LINE_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2)
    assert len(a) == 0


def test_has_line_matching_n_failure():
    """test has_line_matching n .. negative test"""
    a = run_assertions(TEXT_HAS_LINE_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT)
    assert "Expected 2+-0 lines matching for 'te[sx]t te[sx]t' in output ('test text\n') found 1" in a
    assert len(a) == 1


SIZE_HAS_SIZE_ASSERTION = """
    <assert_contents>
        <has_size {size_attrib}="{value}"/>
    </assert_contents>
"""
SIZE_HAS_SIZE_ASSERTION_DELTA = """
    <assert_contents>
        <has_size {size_attrib}="{value}" delta="{delta}"/>
    </assert_contents>
"""

# create gz test data for use with has_size tests
with tempfile.NamedTemporaryFile(mode="w", delete=False) as txttmp:
    txttmp.write("A" * 100)
    txttmp.flush()
    A100 = open(txttmp.name, "rb").read()
    GZA100 = gzip.compress(A100)


@pytest.mark.parametrize("size_attrib", ["size", "value"])
def test_has_size_success(size_attrib):
    """test has_size"""
    a = run_assertions(SIZE_HAS_SIZE_ASSERTION.format(size_attrib=size_attrib, value=10), TEXT_DATA_HAS_TEXT)
    assert len(a) == 0


@pytest.mark.parametrize("size_attrib", ["size", "value"])
def test_has_size_failure(size_attrib):
    """test has_size .. negative test"""
    a = run_assertions(SIZE_HAS_SIZE_ASSERTION.format(size_attrib=size_attrib, value="10"), TEXT_DATA_HAS_TEXT * 2)
    assert "Expected file size of 10+-0 found 20" in a
    assert len(a) == 1


def test_has_size_delta():
    """test has_size .. delta"""
    a = run_assertions(
        SIZE_HAS_SIZE_ASSERTION_DELTA.format(size_attrib="size", value="10", delta="10"), TEXT_DATA_HAS_TEXT * 2
    )
    assert len(a) == 0


def test_has_size_with_bytes_suffix():
    """test has_size .. bytes suffix"""
    a = run_assertions(
        SIZE_HAS_SIZE_ASSERTION_DELTA.format(size_attrib="size", value="1k", delta="0"), TEXT_DATA_HAS_TEXT * 100
    )
    assert len(a) == 0


def test_has_size_with_bytes_suffix_failure():
    """test has_size .. bytes suffix .. negative"""
    a = run_assertions(
        SIZE_HAS_SIZE_ASSERTION_DELTA.format(size_attrib="value", value="1Mi", delta="10k"), TEXT_DATA_HAS_TEXT * 100
    )
    assert "Expected file size of 1Mi+-10k found 1000" in a
    assert len(a) == 1


def test_has_size_decompress_gz():
    """test has_size with gzipped data using decompress=True (which in real life is set int he parent output tag)"""
    a = run_assertions(SIZE_HAS_SIZE_ASSERTION.format(size_attrib="size", value="100"), GZA100, decompress=True)
    assert len(a) == 0


def test_has_size_decompress_txt():
    """
    test has_size with NON-gzipped data using decompress=True
    -> decompress should be ignored - in particular there should be no error
    """
    a = run_assertions(SIZE_HAS_SIZE_ASSERTION.format(size_attrib="size", value="100"), A100, decompress=True)
    assert len(a) == 0


VALID_XML = """<root>
    <elem name="foo">
        <more name="bar">BAR</more>
        <more name="baz">BAZ</more>
        <more name="qux">QUX</more>
    </elem>
    <elem name="baz"/>
</root>
"""

INVALID_XML = '<root><elem name="foo"></root>'

XML_IS_VALID_XML_ASSERTION = """
    <assert_contents>
        <is_valid_xml/>
    </assert_contents>
"""


def test_is_valid_xml_success():
    """test is_valid_xml"""
    a = run_assertions(XML_IS_VALID_XML_ASSERTION, VALID_XML)
    assert len(a) == 0


def test_is_valid_xml_failure():
    """test is_valid_xml .. negative test"""
    a = run_assertions(XML_IS_VALID_XML_ASSERTION, INVALID_XML)
    assert (
        "Expected valid XML, but could not parse output. Opening and ending tag mismatch: elem line 1 and root, line 1, column 31 (<string>, line 1)"
        in a
    )


XML_HAS_ELEMENT_WITH_PATH = """
    <assert_contents>
        <has_element_with_path path="{path}"/>
    </assert_contents>
"""
XML_HAS_N_ELEMENTS_WITH_PATH = """
    <assert_contents>
        <is_valid_xml/>
        <has_n_elements_with_path path="{path}" n="{n}"/>
    </assert_contents>
"""
XML_ELEMENT_TEXT_MATCHES = """
    <assert_contents>
        <element_text_matches path="{path}" expression="{expression}"/>
    </assert_contents>
"""
XML_ELEMENT_TEXT_IS = """
    <assert_contents>
        <element_text_is path="{path}" text="{text}"/>
    </assert_contents>
"""
XML_ATTRIBUTE_MATCHES = """
    <assert_contents>
        <is_valid_xml/>
        <attribute_matches path="{path}" attribute="{attribute}" expression="{expression}"/>
    </assert_contents>
"""
XML_ELEMENT_TEXT = """
    <assert_contents>
        <element_text path="{path}">
            {content_assert}
        </element_text>
    </assert_contents>
"""

XML_XML_ELEMENT = """
    <assert_contents>
        <xml_element path="{path}" attribute="{attribute}" all="{all}" n="{n}" delta="{delta}" min="{min}" max="{max}" negate="{negate}">
            {content_assert}
        </xml_element>
    </assert_contents>
"""


def test_has_element_with_path_success_1():
    """test has_element_with_path"""
    a = run_assertions(XML_HAS_ELEMENT_WITH_PATH.format(path="./elem[1]/more"), VALID_XML)
    assert len(a) == 0


def test_has_element_with_path_success_2():
    """test has_element_with_path"""
    a = run_assertions(XML_HAS_ELEMENT_WITH_PATH.format(path="./elem[@name='foo']"), VALID_XML)
    assert len(a) == 0


def test_has_element_with_path_success_3():
    """test has_element_with_path"""
    a = run_assertions(XML_HAS_ELEMENT_WITH_PATH.format(path=".//more[@name]"), VALID_XML)
    assert len(a) == 0


def test_has_element_with_path_failure():
    """test has_element_with_path .. negative test"""
    a = run_assertions(XML_HAS_ELEMENT_WITH_PATH.format(path="./blah"), VALID_XML)
    assert "Expected path './blah' in xml" in a


def test_has_n_elements_with_path_success_1():
    """test has_n_elements_with_path"""
    a = run_assertions(XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem", n="2"), VALID_XML)
    assert len(a) == 0


def test_has_n_elements_with_path_success_2():
    """test has_n_elements_with_path"""
    a = run_assertions(XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem[1]/more", n="3"), VALID_XML)
    assert len(a) == 0


def test_has_n_elements_with_path_success_3():
    """test has_n_elements_with_path"""
    a = run_assertions(XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem[@name='foo']/more", n="3"), VALID_XML)
    assert len(a) == 0


def test_has_n_elements_with_path_success_4():
    """test has_n_elements_with_path"""
    a = run_assertions(XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem[2]/more", n="0"), VALID_XML)
    assert len(a) == 0


def test_has_n_elements_with_path_failure():
    """test has_n_elements_with_path .. negative test"""
    a = run_assertions(XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem", n="1"), VALID_XML)
    assert "Expected 1+-0 occurrences of path './elem' in xml found 2" in a
    assert len(a) == 1


def test_element_text_matches_sucess():
    """test element_text_matches"""
    a = run_assertions(XML_ELEMENT_TEXT_MATCHES.format(path="./elem/more", expression="BA(R|Z)"), VALID_XML)
    assert len(a) == 0


def test_element_text_matches_sucess_with_more_specific_path():
    """test element_text_matches more specific path"""
    a = run_assertions(XML_ELEMENT_TEXT_MATCHES.format(path="./elem/more[2]", expression="BA(R|Z)"), VALID_XML)
    assert len(a) == 0


def test_element_text_matches_failure():
    """test element_text_matches .. negative test"""
    a = run_assertions(
        XML_ELEMENT_TEXT_MATCHES.format(path="./elem/more", expression="QU(X|Y)"),
        VALID_XML,
    )
    assert "Text of element with path './elem/more': Expected text matching expression 'QU(X|Y)' in output ('BAR')" in a
    assert len(a) == 1


def test_element_text_is_sucess():
    """test element_text_is"""
    a = run_assertions(XML_ELEMENT_TEXT_IS.format(path="./elem/more", text="BAR"), VALID_XML)
    assert len(a) == 0


def test_element_text_is_sucess_with_more_specific_path():
    """test element_text_is with more specific path"""
    a = run_assertions(XML_ELEMENT_TEXT_IS.format(path="./elem/more[@name='baz']", text="BAZ"), VALID_XML)
    assert len(a) == 0


def test_element_text_is_failure():
    """test element_text_is .. negative test testing that prefix is not accepted"""
    a = run_assertions(XML_ELEMENT_TEXT_IS.format(path="./elem/more", text="BA"), VALID_XML)
    assert "Text of element with path './elem/more': Expected text matching expression 'BA$' in output ('BAR')" in a
    assert len(a) == 1


def test_attribute_matches_sucess():
    """est element_attribute_matches"""
    a = run_assertions(
        XML_ATTRIBUTE_MATCHES.format(path="./elem/more", attribute="name", expression="ba(r|z)"), VALID_XML
    )
    assert len(a) == 0


def test_attribute_matches_sucess_with_more_specific_path():
    """test element_attribute_matches with more specific path"""
    a = run_assertions(
        XML_ATTRIBUTE_MATCHES.format(path="./elem/more[2]", attribute="name", expression="ba(r|z)"), VALID_XML
    )
    assert len(a) == 0


def test_attribute_matches_failure():
    """test element_attribute_matches .. negative test"""
    a = run_assertions(
        XML_ATTRIBUTE_MATCHES.format(path="./elem/more", attribute="name", expression="qu(x|y)"), VALID_XML
    )
    assert (
        "Attribute 'name' on element with path './elem/more': Expected text matching expression 'qu(x|y)' in output ('bar')"
        in a
    )
    assert len(a) == 1


def test_element_text_success():
    """test element_text"""
    a = run_assertions(XML_ELEMENT_TEXT.format(path="./elem/more", content_assert=""), VALID_XML)
    assert len(a) == 0


def test_element_text_failure():
    """test element_text .. negative"""
    a = run_assertions(XML_ELEMENT_TEXT.format(path="./absent", content_assert=""), VALID_XML)
    assert "Expected path './absent' in xml" in a
    assert len(a) == 1


def test_element_text_with_subassertion_sucess():
    """test element_text with sub-assertion"""
    a = run_assertions(XML_ELEMENT_TEXT.format(path="./elem/more", content_assert='<has_text text="BAR"/>'), VALID_XML)
    assert len(a) == 0


def test_element_text_with_subassertion_failure():
    """test element_text with sub-assertion .. negative"""
    a = run_assertions(
        XML_ELEMENT_TEXT.format(path="./elem/more", content_assert='<has_text text="NOTBAR"/>'), VALID_XML
    )
    assert "Text of element with path './elem/more': Expected text 'NOTBAR' in output ('BAR')" in a
    assert len(a) == 1


# note that xml_element is also tested indirectly by the other xml
# assertions which are all implemented by xml_element
def test_xml_element_matching_text_success():
    """test xml_element"""
    a = run_assertions(
        XML_XML_ELEMENT.format(
            path=".//more",
            n="2",
            delta="1",
            min="1",
            max="3",
            attribute="",
            all="false",
            content_assert='<has_text_matching expression="(BA[RZ]|QUX)$"/>',
            negate="false",
        ),
        VALID_XML,
    )
    assert len(a) == 0


def test_xml_element_matching_attribute_success():
    """test xml_element testing attribute matching on all matching elements"""
    a = run_assertions(
        XML_XML_ELEMENT.format(
            path=".//more",
            n="2",
            delta="1",
            min="1",
            max="3",
            attribute="name",
            all="true",
            content_assert='<has_text_matching expression="(ba[rz]|qux)$"/>',
            negate="false",
        ),
        VALID_XML,
    )
    assert len(a) == 0


def test_xml_element_failure_due_to_n():
    """test xml_element .. failing because of n"""
    a = run_assertions(
        XML_XML_ELEMENT.format(
            path=".//more",
            n="2",
            delta="0",
            min="1",
            max="3",
            attribute="",
            all="false",
            content_assert="",
            negate="false",
        ),
        VALID_XML,
    )
    assert "Expected 2+-0 occurrences of path './/more' in xml found 3" in a
    assert len(a) == 1


def test_xml_element_failure_due_to_minmax_in_combination_with_negate():
    """test xml_element .. failing because of n"""
    a = run_assertions(
        XML_XML_ELEMENT.format(
            path=".//more",
            n="10000",
            delta="1",
            min="1",
            max="3",
            attribute="",
            all="false",
            content_assert="",
            negate="true",
        ),
        VALID_XML,
    )
    assert "Did not expect that the number of occurences of path './/more' in xml is in [1:3] found 3" in a
    assert len(a) == 1


def test_xml_element_failure_due_to_subassertion():
    """test xml_element .. failing because of sub assertion"""
    a = run_assertions(
        XML_XML_ELEMENT.format(
            path=".//more",
            n="2",
            delta="1",
            min="1",
            max="3",
            attribute="",
            all="false",
            content_assert='<has_text_matching expression="(BA[RZ]|QUX)$" negate="true"/>',
            negate="false",
        ),
        VALID_XML,
    )
    assert (
        "Text of element with path './/more': Did not expect text matching expression '(BA[RZ]|QUX)$' in output ('BAR')"
        in a
    )
    assert len(a) == 1


# create a test directory structure for zipping
# might also be done directly with the fipfile/tarfile module without creating
# a tmpdir, but its much harder to create empty directories or symlinks
tmpdir = tempfile.mkdtemp()
for f in ["file1.txt", "testdir/file1.txt", "testdir/file2.txt", "testdir/dir2/file1.txt"]:
    tmpfile = os.path.join(tmpdir, f)
    os.makedirs(os.path.dirname(tmpfile), exist_ok=True)
    with open(tmpfile, "w") as fh:
        fh.write(f)
os.makedirs(os.path.join(tmpdir, "emptydir"))
os.symlink("testdir/file1.txt", os.path.join(tmpdir, "symlink"))

with tempfile.NamedTemporaryFile(suffix=".zip", delete=False) as ziptmp:
    zipname = ziptmp.name
    shutil.make_archive(zipname[:-4], "zip", tmpdir)
    with open(zipname, "rb") as zfh:
        ZIPBYTES = zfh.read()
with tempfile.NamedTemporaryFile(suffix=".tar.gz", delete=False) as ziptmp:
    zipname = ziptmp.name
    shutil.make_archive(zipname[:-7], "gztar", tmpdir)
    with open(zipname, "rb") as zfh:
        TARBYTES = zfh.read()
shutil.rmtree(tmpdir)


with tempfile.NamedTemporaryFile(mode="w", delete=False) as nonarchivetmp:
    nonarchivename = nonarchivetmp.name
    nonarchivetmp.write("some text")
with open(nonarchivename, "rb") as ntmp:
    NONARCHIVE = ntmp.read()

ARCHIVE_HAS_ARCHIVE_MEMBER = """
    <assert_contents>
        <has_archive_member path="{path}" all="{all}">
            {content_assert}
        </has_archive_member>
    </assert_contents>
"""

ARCHIVE_HAS_ARCHIVE_MEMBER_N = """
    <assert_contents>
        <has_archive_member path="{path}" n="{n}" delta="{delta}">
            {content_assert}
        </has_archive_member>
    </assert_contents>
"""

ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX = """
    <assert_contents>
        <has_archive_member path="{path}" min="{min}" max="{max}">
            {content_assert}
        </has_archive_member>
    </assert_contents>
"""


def test_has_archive_member_1filegzip():
    """test has_archive_member with a single file gz which should fail (has no members anyway)"""
    a = run_assertions(ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?xyz", content_assert="", all="false"), GZA100)
    assert "Expected path '(\\./)?xyz' to be an archive" in a
    assert len(a) == 1


def test_has_archive_member_zip():
    """test has_archive_member with zip"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/file1.txt", content_assert="", all="false"), ZIPBYTES
    )
    assert len(a) == 0


def test_has_archive_member_tar():
    """test has_archive_member with tar"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/file1.txt", content_assert="", all="false"), TARBYTES
    )
    assert len(a) == 0


def test_has_archive_member_nonarchive():
    """test has_archive_member with non archive"""
    a = run_assertions(ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="irrelevant", content_assert="", all="false"), NONARCHIVE)
    assert "Expected path 'irrelevant' to be an archive" in a
    assert len(a) == 1


def test_has_archive_member_zip_absent_member():
    """test has_archive_member with zip on absent member"""
    a = run_assertions(ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="absent", content_assert="", all="false"), ZIPBYTES)
    assert "Expected path 'absent' in archive" in a
    assert len(a) == 1


def test_has_archive_member_tar_absent_member():
    """test has_archive_member with tar on absent member"""
    a = run_assertions(ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="absent", content_assert="", all="false"), TARBYTES)
    assert "Expected path 'absent' in archive" in a
    assert len(a) == 1


def test_has_archive_member_zip_symlink_member():
    """test has_archive_member with zip on symlink"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path="(\\./)?symlink", content_assert='<has_text text="testdir/file1.txt"/>', all="false"
        ),
        ZIPBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_tar_symlink_member():
    """test has_archive_member with tar on symlink"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path="(\\./)?symlink", content_assert='<has_text text="testdir/file1.txt"/>', all="false"
        ),
        TARBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_zip_nonfile_member():
    """test has_archive_member with zip on a dir member (which are treated like empty files)"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/", content_assert='<has_size value="0"/>', all="false"),
        ZIPBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_tar_nonfile_member():
    """test has_archive_member with tar on a dir member (which are treated like empty files)"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/", content_assert='<has_size value="0"/>', all="false"),
        TARBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_zip_with_content_assertion():
    """test has_archive_member with zip with subassertion (note that archive members are sorted therefor file1 in dir2 is tested)"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path="(\\./)?testdir/.*\\.txt", content_assert='<has_text text="testdir/dir2/file1.txt"/>', all="false"
        ),
        ZIPBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_tar_with_content_assertion():
    """test has_archive_member with tar with subassertion (note that archive members are sorted therefor file1 in dir2 is tested)"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path="(\\./)?testdir/.*\\.txt", content_assert='<has_text text="testdir/dir2/file1.txt"/>', all="false"
        ),
        TARBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_zip_with_failing_content_assertion():
    """test has_archive_member with zip with failing subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path="(\\./)?testdir/file1.txt", content_assert='<has_text text="ABSENT"/>', all="false"
        ),
        ZIPBYTES,
    )
    assert "Archive member '(\\./)?testdir/file1.txt': Expected text 'ABSENT' in output ('testdir/file1.txt')" in a
    assert len(a) == 1


def test_has_archive_member_tar_with_failing_content_assertion():
    """test has_archive_member with tar with failing subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path="(\\./)?testdir/file1.txt", content_assert='<has_text text="ABSENT"/>', all="false"
        ),
        TARBYTES,
    )
    assert "Archive member '(\\./)?testdir/file1.txt': Expected text 'ABSENT' in output ('testdir/file1.txt')" in a
    assert len(a) == 1


def test_has_archive_member_zip_all_matching_with_content_assertion():
    """test has_archive_member with zip checking all matches with subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', all="true"
        ),
        ZIPBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_tar_all_matching_with_content_assertion():
    """test has_archive_member with tar checking all matches with subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', all="true"
        ),
        TARBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_zip_all_matching_with_failing_content_assertion():
    """test has_archive_member with zip checking all matches with failing subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file1\\.txt"/>', all="true"
        ),
        ZIPBYTES,
    )
    assert "Expected text matching expression 'file1\\.txt' in output ('testdir/file2.txt')"
    assert len(a) == 1


def test_has_archive_member_tar_all_matching_with_failing_content_assertion():
    """test has_archive_member with tar checking all matches with failing subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file1\\.txt"/>', all="true"
        ),
        TARBYTES,
    )

    assert (
        "Archive member '.*file.\\.txt': Expected text matching expression 'file1\\.txt' in output ('testdir/file2.txt')"
        in a
    )
    assert len(a) == 1


def test_has_archive_member_zip_n_delta_and_content_assertion():
    """test has_archive_member with zip n+delta with subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="3", delta="1"
        ),
        ZIPBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_zip_n_delta_failing_and_content_assertion():
    """test has_archive_member with zip n+delta with subassertion .. negative"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="1", delta="1"
        ),
        ZIPBYTES,
    )
    assert "Expected 1+-1 matches for path '.*file.\\.txt' in archive found 4" in a
    assert len(a) == 1


def test_has_archive_member_tar_n_delta_and_content_assertion():
    """test has_archive_member with tar n+delta with subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="3", delta="1"
        ),
        TARBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_tar_n_delta_failing_and_content_assertion():
    """test has_archive_member with tar n+delta with subassertion .. negative"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="1", delta="1"
        ),
        TARBYTES,
    )
    assert "Expected 1+-1 matches for path '.*file.\\.txt' in archive found 4" in a
    assert len(a) == 1


def test_has_archive_member_zip_min_max_and_content_assertion():
    """test has_archive_member with zip min+max with subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="2", max="4"
        ),
        ZIPBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_zip_min_max_failing_and_content_assertion():
    """test has_archive_member with zip min+max with subassertion .. negative"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="0", max="2"
        ),
        ZIPBYTES,
    )
    assert "Expected that the number of matches for path '.*file.\\.txt' in archive is in [0:2] found 4" in a
    assert len(a) == 1


def test_has_archive_member_tar_min_max_and_content_assertion():
    """test has_archive_member with tar min+max with subassertion"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="2", max="4"
        ),
        TARBYTES,
    )
    assert len(a) == 0


def test_has_archive_member_tar_min_max_failing_and_content_assertion():
    """test has_archive_member with tar min+max with subassertion .. negative"""
    a = run_assertions(
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(
            path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="0", max="2"
        ),
        TARBYTES,
    )
    assert "Expected that the number of matches for path '.*file.\\.txt' in archive is in [0:2] found 4" in a
    assert len(a) == 1


JSON_HAS_PROPERTY_WITH_VALUE = """
    <assert_contents>
        <has_json_property_with_value property="{property}" value="{value}" />
    </assert_contents>
"""

JSON_HAS_PROPERTY_WITH_TEXT = """
    <assert_contents>
        <has_json_property_with_text property="{property}" text="{text}" />
    </assert_contents>
"""

VALID_SIMPLE_JSON = """{"foo": 5, "list": [{"textprop": "right"}]}"""


def test_has_json_property_with_value_pos():
    """positive test for has_json_property_with_value"""
    a = run_assertions(
        JSON_HAS_PROPERTY_WITH_VALUE.format(property="foo", value="5"),
        VALID_SIMPLE_JSON,
    )
    assert len(a) == 0


def test_has_json_property_with_value_neg():
    """negative test for has_json_property_with_value"""
    a = run_assertions(
        JSON_HAS_PROPERTY_WITH_VALUE.format(property="foo", value="6"),
        VALID_SIMPLE_JSON,
    )
    assert "Failed to find property [foo] with JSON value [6]" in a
    assert len(a) == 1


def test_has_json_property_with_text_pos():
    """positive test for has_json_property_with_text"""
    a = run_assertions(
        JSON_HAS_PROPERTY_WITH_TEXT.format(property="textprop", text="right"),
        VALID_SIMPLE_JSON,
    )
    assert len(a) == 0


def test_has_json_property_with_text_neg():
    """negative test for has_json_property_with_text"""
    a = run_assertions(
        JSON_HAS_PROPERTY_WITH_TEXT.format(property="textprop", text="wrong"),
        VALID_SIMPLE_JSON,
    )
    assert "Failed to find property [textprop] with text [wrong]" in a
    assert len(a) == 1


if h5py is not None:
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        h5name = tmp.name
        with h5py.File(tmp.name, "w", locking=False) as h5fh:
            h5fh.attrs["myfileattr"] = "myfileattrvalue"
            h5fh.attrs["myfileattrint"] = 1
            dset = h5fh.create_dataset("myint", (100,), dtype="i")
            dset.attrs["myintattr"] = "myintattrvalue"
            grp = h5fh.create_group("mygroup")
            grp.attrs["mygroupattr"] = "mygroupattrvalue"
            grp.create_dataset("myfloat", (50,), dtype="f")
            dset.attrs["myfloatattr"] = "myfloatattrvalue"
    with open(h5name, "rb") as h5fh:
        H5BYTES = h5fh.read()
    os.remove(h5name)

    H5_HAS_H5_KEYS = """
        <assert_contents>
            <has_h5_keys keys="myint,mygroup,mygroup/myfloat"/>
        </assert_contents>
    """
    H5_HAS_H5_KEYS_NEGATIVE = """
        <assert_contents>
            <has_h5_keys keys="absent"/>
        </assert_contents>
    """
    H5_HAS_ATTRIBUTE = """
        <assert_contents>
            <has_h5_attribute key="myfileattr" value="myfileattrvalue" />
            <has_h5_attribute key="myfileattrint" value="1" />
        </assert_contents>
    """
    H5_HAS_ATTRIBUTE_NEGATIVE = """
        <assert_contents>
            <has_h5_attribute key="myfileattr" value="wrong" />
            <has_h5_attribute key="myfileattrint" value="also_wrong" />
        </assert_contents>
    """

    def test_has_h5_keys():
        """test has_h5_keys"""
        a = run_assertions(H5_HAS_H5_KEYS, H5BYTES)
        assert len(a) == 0

    def test_has_h5_keys_failure():
        """test has_h5_keys .. negative"""
        a = run_assertions(H5_HAS_H5_KEYS_NEGATIVE, H5BYTES)
        assert "Not a HDF5 file or H5 keys missing:\n\t['mygroup', 'mygroup/myfloat', 'myint']\n\t['absent']" in a
        assert len(a) == 1

    def test_has_h5_attribute():
        """test has_attribut"""
        a = run_assertions(H5_HAS_ATTRIBUTE, H5BYTES)
        assert len(a) == 0

    def test_has_h5_attribute_failure():
        """test has_attribute .. negative"""
        a = run_assertions(H5_HAS_ATTRIBUTE_NEGATIVE, H5BYTES)
        assert (
            "Not a HDF5 file or H5 attributes do not match:\n\t[('myfileattr', 'myfileattrvalue'), ('myfileattrint', '1')]\n\n\t(myfileattr : wrong)"
            in a
        )
        assert len(a) == 1


def run_assertions(assertion_xml: str, data, decompress=False) -> Tuple:
    assertion = parse_xml_string(assertion_xml)
    assertion_description = __parse_assert_list_from_elem(assertion)
    assert assertion_description
    try:
        asserts.verify_assertions(data, assertion_description, decompress=decompress)
    except AssertionError as e:
        assert_list = e.args
    else:
        assert_list = ()
    return assert_list



import pytest
from packaging.version import Version

from galaxy.tool_util.version import (
    LegacyVersion,
    parse_version,
)

# This list must be in the correct sorting order
VERSIONS = [
    # Implicit epoch of 0
    "1.0.dev456",
    "1.0a1",
    "1.0a2.dev456",
    "1.0a12.dev456",
    "1.0a12",
    "1.0b1.dev456",
    "1.0b2",
    "1.0b2.post345.dev456",
    "1.0b2.post345",
    "1.0b2-346",
    "1.0c1.dev456",
    "1.0c1",
    "1.0rc2",
    "1.0c3",
    "1.0",
    "1.0.post456.dev34",
    "1.0.post456",
    "1.1.dev1",
    "1.2+123abc",
    "1.2+123abc456",
    "1.2+abc",
    "1.2+abc123",
    "1.2+abc123def",
    "1.2+1234.abc",
    "1.2+123456",
    "1.2.r32+123456",
    "1.2.rev33+123456",
    # Explicit epoch of 1
    "1!1.0.dev456",
    "1!1.0a1",
    "1!1.0a2.dev456",
    "1!1.0a12.dev456",
    "1!1.0a12",
    "1!1.0b1.dev456",
    "1!1.0b2",
    "1!1.0b2.post345.dev456",
    "1!1.0b2.post345",
    "1!1.0b2-346",
    "1!1.0c1.dev456",
    "1!1.0c1",
    "1!1.0rc2",
    "1!1.0c3",
    "1!1.0",
    "1!1.0.post456.dev34",
    "1!1.0.post456",
    "1!1.1.dev1",
    "1!1.2+123abc",
    "1!1.2+123abc456",
    "1!1.2+abc",
    "1!1.2+abc123",
    "1!1.2+abc123def",
    "1!1.2+1234.abc",
    "1!1.2+123456",
    "1!1.2.r32+123456",
    "1!1.2.rev33+123456",
]

LEGACY_VERSIONS = ["foobar", "a cat is fine too", "lolwut", "1.8.1+2+galaxy0"]


@pytest.mark.parametrize("version", VERSIONS)
def test_parse_pep440_versions(version: str) -> None:
    assert isinstance(parse_version(version), Version)


@pytest.mark.parametrize("version", LEGACY_VERSIONS)
def test_legacy_version_fields(version: str) -> None:
    parsed_version = parse_version(version)
    assert isinstance(parsed_version, LegacyVersion)
    assert parsed_version.public == version
    assert parsed_version.base_version == version
    assert parsed_version.epoch == -1
    assert parsed_version.release is None
    assert parsed_version.pre is None
    assert parsed_version.post is None
    assert parsed_version.dev is None
    assert parsed_version.local is None
    assert parsed_version.is_prerelease is False
    assert parsed_version.is_postrelease is False
    assert parsed_version.is_devrelease is False


@pytest.mark.parametrize(
    "lower_ver,greater_ver",
    [(VERSIONS[i], VERSIONS[i + 1]) for i in range(len(VERSIONS) - 1)]
    + [(lv, v) for v in VERSIONS for lv in LEGACY_VERSIONS],
)
def test_version_cmp(lower_ver: str, greater_ver: str) -> None:
    assert parse_version(lower_ver) < parse_version(greater_ver)



from galaxy.tool_util.biotools.source import (
    ApiBiotoolsMetadataSource,
    BiotoolsMetadataSourceConfig,
    get_biotools_metadata_source,
    GitContentBiotoolsMetadataSource,
)
from galaxy.util.unittest_utils import skip_if_site_down
from ._util import content_dir


def test_git_content():
    metadata_source = GitContentBiotoolsMetadataSource(content_dir)
    bwa_entry = metadata_source.get_biotools_metadata("bwa")
    assert bwa_entry is not None
    assert len(bwa_entry.function) == 6

    missing_entry = metadata_source.get_biotools_metadata("johnscoolbowtie")
    assert missing_entry is None


@skip_if_site_down("https://bio.tools/")
def test_api_content():
    metadata_source = ApiBiotoolsMetadataSource()
    bwa_entry = metadata_source.get_biotools_metadata("bwa")
    assert bwa_entry is not None
    assert len(bwa_entry.function) >= 6

    missing_entry = metadata_source.get_biotools_metadata("johnscoolbowtie")
    assert missing_entry is None


@skip_if_site_down("https://bio.tools/")
def test_cascade_content():
    config = BiotoolsMetadataSourceConfig()
    config.content_directory = content_dir
    metadata_source = get_biotools_metadata_source(config)
    bwa_entry = metadata_source.get_biotools_metadata("bwa")
    assert bwa_entry is not None
    assert len(bwa_entry.function) == 6

    # by default API isn't used so bowtie2 empty...
    bowtie2 = metadata_source.get_biotools_metadata("bowtie2")
    assert bowtie2 is None

    # but can be enabled with API
    config = BiotoolsMetadataSourceConfig()
    config.content_directory = content_dir
    config.use_api = True
    metadata_source = get_biotools_metadata_source(config)
    bowtie2 = metadata_source.get_biotools_metadata("bowtie2")
    assert bowtie2 is not None
    assert len(bowtie2.topic) == 3
